#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# pylint: disable=line-too-long

import os
import sys
import unittest
import tempfile
import uuid
import hmac
from time import monotonic
from typing import Callable

from xpra.os_util import WIN32, OSX, POSIX, get_hex_uuid
from xpra.util.env import OSEnvContext
from xpra.util.str_fn import strtobytes
from xpra.util.objects import typedict
from xpra.net.digest import get_digests, get_digest_module, gendigest, get_salt


def temp_filename(prefix="") -> str:
    return os.path.join(tempfile.gettempdir(), "file-auth-%s-test-%s" % (prefix, monotonic()))


class TempFileContext:

    def __init__(self, prefix="prefix"):
        self.prefix = prefix
        self.filename = ""
        self.file = None

    def __enter__(self):
        if WIN32:
            # NamedTemporaryFile doesn't work for reading on win32...
            self.filename = temp_filename(self.prefix)
            self.file = open(self.filename, 'w')
        else:
            self.file = tempfile.NamedTemporaryFile(mode="w", prefix=self.prefix)
            self.filename = self.file.name
        return self

    def __exit__(self, _exc_type, _exc_val, _exc_tb):
        if WIN32 and self.filename:
            os.unlink(self.filename)


class TestAuth(unittest.TestCase):

    def a(self, name: str):
        pmod = "xpra.auth"
        auth_module = __import__(pmod, globals(), locals(), [name], 0)
        mod = getattr(auth_module, name, None)
        assert mod, f"cannot load {name} from {pmod}"
        assert str(mod)
        return mod

    def _init_auth(self, mod_name: str, **kwargs):
        mod = self.a(mod_name)
        a = self.do_init_auth(mod, **kwargs)
        assert repr(a)
        return a

    def do_init_auth(self, module, **kwargs):
        try:
            c = module.Authenticator
        except AttributeError:
            raise Exception("module %s does not contain an Authenticator class!") from None
        # some auth modules require this to function:
        if "connection" not in kwargs:
            kwargs["connection"] = "fake-connection-data"
        # exec auth would fail during rpmbuild without a default command:
        if "command" not in kwargs:
            kwargs["command"] = "/bin/true"
        kwargs["username"] = kwargs.get("username", "foo")
        return c(**kwargs)

    def _test_module(self, module):
        a = self._init_auth(module)
        assert a
        assert str(a)
        assert repr(a)
        if a.requires_challenge():
            salt, digest = a.get_challenge(get_digests())
            assert salt and digest
        a = self._init_auth(module)
        assert a
        if a.requires_challenge():
            salt = digest = ""
            try:
                salt, digest = a.get_challenge(("invalid-digest",))
            except ValueError:
                pass
            else:
                raise RuntimeError(f"invalid digest should raise a ValueError, but got ({salt}, {digest})")

    def capsauth(self, a, challenge_response=None, client_salt=None) -> bool:
        caps = typedict()
        if challenge_response is not None:
            caps["challenge_response"] = challenge_response
        if client_salt is not None:
            caps["challenge_client_salt"] = client_salt
        return a.authenticate(caps)

    def test_all(self) -> None:
        test_modules = ["reject", "allow", "none", "file", "multifile", "env", "password"]
        try:
            self.a("pam")
            test_modules.append("pam")
        except Exception:
            pass
        if sys.platform.startswith("win"):
            self.a("win32")
            test_modules.append("win32")
        if POSIX:
            test_modules.append("exec")
        for module in test_modules:
            self._test_module(module)

    def test_fail(self) -> None:
        try:
            fa = self._init_auth("fail")
        except Exception:
            fa = None
        assert fa is None, "'fail_auth' did not fail!"

    def test_reject(self) -> None:
        a = self._init_auth("reject")
        assert a.requires_challenge()
        c, mac = a.get_challenge(get_digests())
        assert a.get_uid() == -1
        assert a.get_gid() == -1
        assert not a.get_password()
        assert c and mac
        assert not a.get_sessions()
        assert not a.get_passwords()
        assert a.choose_salt_digest("xor") == "xor"
        for x in (None, "bar"):
            assert not self.capsauth(a, x, c)
            assert not self.capsauth(a, x, x)

    def test_none(self) -> None:
        a = self._init_auth("none")
        assert not a.requires_challenge()
        salt, digest = a.get_challenge(get_digests())
        assert not (salt or digest)
        assert not a.get_password()
        for x in (None, "bar"):
            assert self.capsauth(a, x, "")
            assert self.capsauth(a, "", x)

    def test_allow(self) -> None:
        a = self._init_auth("allow")
        assert a.requires_challenge()
        assert a.get_challenge(get_digests())
        assert not a.get_passwords()
        for x in (None, "bar"):
            assert self.capsauth(a, x, "")
            assert self.capsauth(a, "", x)

    def _test_hmac_auth(self, mod_name: str, password: str, **kwargs) -> None:
        for test_password in (password, "somethingelse"):
            a = self._init_auth(mod_name, **kwargs)
            assert a.requires_challenge()
            assert a.get_passwords()
            salt, mac = a.get_challenge(tuple(x for x in get_digests() if x.startswith("hmac")))
            assert salt
            assert mac.startswith("hmac"), "invalid mac: %s" % mac
            client_salt = strtobytes(uuid.uuid4().hex + uuid.uuid4().hex)
            salt_digest = a.choose_salt_digest(get_digests())
            auth_salt = strtobytes(gendigest(salt_digest, client_salt, salt))
            digestmod = get_digest_module(mac)
            verify = hmac.HMAC(strtobytes(test_password), auth_salt, digestmod=digestmod).hexdigest()
            passed = self.capsauth(a, verify, client_salt)
            assert passed == (test_password == password), "expected authentication to %s with %s vs %s" % (
                ["fail", "succeed"][test_password == password], test_password, password)
            assert not self.capsauth(a, verify,
                                     client_salt), "should not be able to athenticate again with the same values"

    def test_env(self) -> None:
        for var_name in ("XPRA_PASSWORD", "SOME_OTHER_VAR_NAME"):
            password = uuid.uuid4().hex
            os.environ[var_name] = password
            try:
                kwargs = {}
                if var_name != "XPRA_PASSWORD":
                    kwargs["name"] = var_name
                self._test_hmac_auth("env", password, name=var_name)
            finally:
                del os.environ[var_name]

    def test_password(self) -> None:
        password = uuid.uuid4().hex
        self._test_hmac_auth("password", password, value=password)

    def _test_file_auth(self, mod_name: str, genauthdata: Callable, display_count: int = 0):
        # no file, no go:
        a = self._init_auth(mod_name)
        assert a.requires_challenge()
        p = a.get_passwords()
        assert not p, "got passwords from %s: %s" % (a, p)
        # challenge twice is a fail
        salt, digest = a.get_challenge(get_digests())
        assert salt and digest
        salt, digest = a.get_challenge(get_digests())
        assert not (salt or digest)
        salt, digest = a.get_challenge(get_digests())
        assert not (salt or digest)
        # muck:
        # 0 - OK
        # 1 - bad: with warning about newline
        # 2 - verify bad passwords
        # 3 - verify no password
        for muck in (0, 1, 2, 3):
            with TempFileContext(prefix=mod_name) as context:
                f = context.file
                filename = context.filename
                with f:
                    a = self._init_auth(mod_name, filename=filename)
                    password, filedata = genauthdata(a)
                    if muck != 3:
                        f.write(filedata)
                    if muck == 1:
                        f.write("\n")
                    f.flush()
                    assert a.requires_challenge()
                    server_salt, digest = a.get_challenge(get_digests())
                    assert get_digest_module(digest)
                    assert server_salt
                    assert digest in get_digests()
                    assert digest != "xor"
                    # this is what a client does,
                    # see send_challenge_reply in client_base
                    client_salt = get_salt(len(server_salt))
                    salt_digest = a.choose_salt_digest(get_digests())
                    assert salt_digest and isinstance(salt_digest, str)
                    auth_salt = gendigest(salt_digest, client_salt, server_salt)
                    assert isinstance(auth_salt,
                                      bytes), f"auth salt is {type(auth_salt)}, expected bytes for {salt_digest!r}"
                    if muck == 0:
                        verify = gendigest(digest, password, auth_salt)
                        assert isinstance(verify, bytes), f"value is {type(verify)}, expected bytes for {digest!r}"
                        assert self.capsauth(a, verify, client_salt), "%s failed" % a.authenticate
                        if display_count > 0:
                            sessions = a.get_sessions()
                            assert len(sessions) >= 3
                            displays = sessions[2]
                            assert len(displays) == display_count, "expected %i displays but got %i : %s" % (
                                display_count, len(sessions), sessions)
                        assert not self.capsauth(a, verify, client_salt), "authenticated twice!"
                        passwords = a.get_passwords()
                        assert len(passwords) == 1, f"expected just one password in file, got {len(passwords)}"
                        assert password in passwords, f"expected to find {password} in {passwords}"
                    else:
                        for verify in ("whatever", None, "bad"):
                            assert not self.capsauth(a, verify, client_salt)
        return a

    def test_file(self) -> None:

        def genfiledata(_a) -> tuple[str, str]:
            password = uuid.uuid4().hex
            return password, password

        self._test_file_auth("file", genfiledata)
        # no digest -> no challenge
        a = self._init_auth("file", filename="foo")
        assert a.requires_challenge()
        try:
            a.get_challenge(("not-a-valid-digest",))
        except ValueError:
            pass
        else:
            raise RuntimeError("invalid digest should raise a ValueError")
        a.password_filename = "./this-path-should-not-exist"
        assert not a.load_password_file()
        assert a.stat_password_filetime() == 0
        # inaccessible:
        if POSIX:
            filename = "./test-file-auth-%s-%s" % (get_hex_uuid(), os.getpid())
            with open(filename, 'wb') as f:
                os.fchmod(f.fileno(), 0o200)  #write-only
            a.password_filename = filename
            a.load_password_file()

    def test_multifile(self):
        def genfiledata(a) -> tuple[str, str]:
            password = uuid.uuid4().hex
            lines = [
                "#comment",
                "%s|%s|||" % (a.username, password),
                "incompleteline",
                "duplicateentry|pass1",
                "duplicateentry|pass2",
                "user|pass",
                "otheruser|otherpassword|1000|1000||env1=A,env2=B|compression=0",
            ]
            return password, "\n".join(lines)

        self._test_file_auth("multifile", genfiledata, 1)

        def nodata(_a) -> tuple[str, str]:
            return "abc", ""

        try:
            self._test_file_auth("multifile", nodata, 1)
        except AssertionError:
            pass
        else:
            raise Exception("authentication with no data should have failed")

    def test_sqlite(self):
        from xpra.auth.sqlite import main as sqlite_main
        filename = temp_filename("sqlite")
        password = "hello"

        def t():
            self._test_hmac_auth("sqlite", password, filename=filename)

        def vf(reason):
            try:
                t()
            except Exception:
                pass
            else:
                raise Exception("sqlite auth should have failed: %s" % reason)

        vf("the database has not been created yet")
        assert sqlite_main(["main", filename, "create"]) == 0
        vf("the user has not been added yet")
        assert sqlite_main(["main", filename, "add", "foo", password]) == 0
        t()
        assert sqlite_main(["main", filename, "remove", "foo"]) == 0
        vf("the user has been removed")
        assert sqlite_main(["main", filename, "add", "foo", "wrongpassword"]) == 0
        vf("the password should not match")

    def test_peercred(self) -> None:
        if not POSIX or OSX:
            # can't be used!
            return
        # no connection supplied:
        pc = self._init_auth("peercred")
        assert not pc.requires_challenge()
        assert not self.capsauth(pc)
        assert pc.get_uid() == -1 and pc.get_gid() == -1
        # now with a connection object:
        from xpra.util.thread import start_thread
        sockpath = "./socket-test"
        try:
            os.unlink(sockpath)
        except OSError:
            pass
        from xpra.net.bytestreams import SocketConnection
        import socket
        sock = socket.socket(socket.AF_UNIX)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind(sockpath)
        sock.listen(5)
        verified = []
        to_close: list[Callable] = [sock.close]

        def wait_for_connection() -> None:
            conn, addr = sock.accept()
            s = SocketConnection(conn, sockpath, addr, sockpath, "unix")
            pc = self._init_auth("peercred", connection=s)
            assert not pc.requires_challenge()
            assert pc.get_uid() == os.getuid()
            verified.append(True)
            to_close.append(s.close)

        t = start_thread(wait_for_connection, "socket listener", daemon=True)
        # connect a client:
        client = socket.socket(socket.AF_UNIX)
        client.settimeout(5)
        client.connect(sockpath)
        to_close.append(client.close)
        # wait for it to trigger auth:
        t.join(5)
        for close in to_close:
            try:
                close()
            except OSError:
                pass
        assert verified

    def test_hosts(self) -> None:
        # cannot be tested (would require root to edit the hosts.deny file)
        pass

    def test_exec(self) -> None:
        if not POSIX:
            return

        def exec_cmd(cmd: str, success=True) -> None:
            kwargs = {
                "command": cmd,
                "timeout": 2,
            }
            a = self._init_auth("exec", **kwargs)
            assert not a.requires_challenge(), f"{a} should not require a challenge"
            expected = ["failed", "succeeded"][success]
            assert self.capsauth(a) == success, f"{a} should have {expected} using cmd={cmd}"

        exec_cmd("/bin/true", True)
        exec_cmd("/bin/false", False)

    def test_keycloak(self) -> None:
        try:
            self._init_auth("keycloak")
            import oauthlib
        except ImportError as e:
            print("Warning: keycloak auth test skipped")
            print(f" {e}")
            return

        def t(digests=None, response=None, **kwargs):
            a = self._init_auth("keycloak", **kwargs)
            assert a.requires_challenge(), "%s should require a challenge" % a
            if digests is not None:
                salt, digest = a.get_challenge(digests)
                assert salt and digest, "cannot get challenge for digests %s" % (digests,)
            if response is not None:
                assert a.check(response), "check failed for response %s" % (response,)

        def f(digests=None, response=None, **kwargs):
            try:
                t(digests, response, **kwargs)
            except (AssertionError, ValueError, oauthlib.oauth2.rfc6749.errors.OAuth2Error, TypeError):
                pass
            else:
                raise Exception("keycloak auth should have failed with arguments: %s" % (kwargs,))

        t()
        # only 'authorization_code' is supported:
        f(grant_type="foo")
        t(grant_type="authorization_code")
        # only 'keycloak' digest is supported:
        f(digests=("xor",))
        t(digests=("xor", "keycloak",))
        t(digests=("keycloak",))
        # we can't provide a valid response:
        # these are not valid json strings:
        for invalid in (True, False, 10, 1.1, [1, 2, 3], (4, 5, 6)):
            f(digests=("keycloak",), response=invalid)
        # these are valid json strings, but not valid reponses (not dicts for a start):
        for invalid in (b"\"hello\"", "\"foo\"", "\"foobar\""):
            f(digests=("keycloak",), response=invalid)
        # these are valid json strings that return a dict, but no valid authorization code:
        f(digests=("keycloak",), response="{\"foo\":\"bar\"}")
        f(digests=("keycloak",), response="{\"error\": 404, \"code\":\"authorization_code\"}")
        f(digests=("keycloak",), response="{\"code\":\"authorization_code\"}")
        # non-https URL should fail:
        f(server_url="http://localhost:8080/")
        with OSEnvContext(OAUTHLIB_INSECURE_TRANSPORT="1"):
            # and succeed with insecure override:
            t(server_url="http://localhost:8080/")


def main():
    import logging
    from xpra.log import set_default_level
    from xpra.log import enable_color
    enable_color()
    if "-v" in sys.argv:
        set_default_level(logging.DEBUG)
    else:
        set_default_level(logging.CRITICAL)
    try:
        from xpra import auth
        assert auth
    except ImportError as e:
        print("non server build, skipping auth module test: %s" % e)
        return
    unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from queue import SimpleQueue
from threading import Timer, RLock
from typing import Any, TypeAlias
from collections.abc import Callable, Sequence

from xpra.util.objects import AtomicInteger
from xpra.log import Logger

log = Logger("util")

ScheduledItemType: TypeAlias = tuple[Callable, Sequence[Any], dict[str, Any]]


# emulate the glib main loop using a single thread + queue:
class QueueScheduler:
    __slots__ = ("main_queue", "exit", "timer_id", "timers", "timer_lock")

    def __init__(self):
        self.main_queue: SimpleQueue[ScheduledItemType | None] = SimpleQueue()
        self.exit = False
        self.timer_id = AtomicInteger()
        self.timers: dict[int, Timer | None] = {}
        self.timer_lock = RLock()

    def source_remove(self, tid: int) -> None:
        log("source_remove(%i)", tid)
        with self.timer_lock:
            timer = self.timers.pop(tid, None)
            if timer:
                timer.cancel()

    def idle_add(self, fn: Callable, *args, **kwargs) -> int:
        tid = self.timer_id.increase()
        self.main_queue.put((self.idle_repeat_call, (tid, fn, args, kwargs), {}))
        # add an entry,
        # but use the value None to stop us from trying to call cancel()
        self.timers[tid] = None
        return tid

    def idle_repeat_call(self, tid: int, fn: Callable, args, kwargs):
        if tid not in self.timers:
            return False  # cancelled
        return fn(*args, **kwargs)

    def timeout_add(self, timeout: int, fn: Callable, *args, **kwargs) -> int:
        tid = self.timer_id.increase()
        self.do_timeout_add(tid, timeout, fn, *args, **kwargs)
        return tid

    def do_timeout_add(self, tid: int, timeout: int, fn: Callable, *args, **kwargs) -> None:
        # emulate glib's timeout_add using Timers
        args = (tid, timeout, fn, args, kwargs)
        t = Timer(timeout / 1000.0, self.queue_timeout_function, args)
        self.timers[tid] = t
        t.start()

    def queue_timeout_function(self, tid: int, timeout: int, fn: Callable, fn_args, fn_kwargs) -> None:
        if tid not in self.timers:  # pragma: no cover
            return  # cancelled
        # add to run queue:
        mqargs = (tid, timeout, fn, fn_args, fn_kwargs)
        self.main_queue.put((self.timeout_repeat_call, mqargs, {}))

    def timeout_repeat_call(self, tid: int, timeout: int, fn: Callable, fn_args, fn_kwargs) -> bool:
        # executes the function then re-schedules it (if it returns True)
        if tid not in self.timers:  # pragma: no cover
            return False  # cancelled
        v = fn(*fn_args, **fn_kwargs)
        if bool(v):
            # create a new timer with the same tid:
            with self.timer_lock:
                if tid in self.timers:
                    self.do_timeout_add(tid, timeout, fn, *fn_args, **fn_kwargs)
        else:
            self.timers.pop(tid, None)
        # we do the scheduling via timers, so always return False here
        # so that the main queue won't re-schedule this function call itself:
        return False

    def run(self) -> None:
        log("run() queue has %s items already in it", self.main_queue.qsize())
        # process "idle_add"/"timeout_add" events in the main loop:
        while not self.exit:
            log("run() size=%s", self.main_queue.qsize())
            v = self.main_queue.get()
            if v is None:
                log("run() None exit marker")
                break
            fn, args, kwargs = v
            log("run() %s%s%s", fn, args, kwargs)
            with log.trap_error(f"Error during main loop callback {fn}"):
                r = fn(*args, **kwargs)
                if bool(r):
                    # re-run it
                    self.main_queue.put(v)
        self.exit = True

    def stop(self) -> None:
        self.exit = True
        self.stop_main_queue()

    def stop_main_queue(self) -> None:
        self.main_queue.put(None)
        # empty the main queue:
        q: SimpleQueue = SimpleQueue()
        q.put(None)
        self.main_queue = q



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from ctypes import (
    byref, addressof,
    FormatError, GetLastError,  # @UnresolvedImport
)
from ctypes.wintypes import HANDLE

from xpra.util.env import envbool
from xpra.platform.win32.common import CloseHandle, LogonUser
from xpra.log import Logger

log = Logger("win32", "auth")

LOG_CREDENTIALS = envbool("XPRA_LOG_CREDENTIALS", False)

MAX_COMPUTERNAME_LENGTH = 15
LOGON32_LOGON_NETWORK_CLEARTEXT = 8
LOGON32_PROVIDER_DEFAULT = 0


def check(domain="", username: str = "", password: str = "") -> bool:
    token = HANDLE()
    # domain = os.environ.get('COMPUTERNAME')
    if LOG_CREDENTIALS:
        log("LogonUser(%s, %s, %s, CLEARTEXT, DEFAULT, %#x)",
            username, domain, password, addressof(token))
    status = LogonUser(username, domain, password,
                       LOGON32_LOGON_NETWORK_CLEARTEXT,
                       LOGON32_PROVIDER_DEFAULT,
                       byref(token))
    log("LogonUser(..)=%#x", status)
    if status:
        CloseHandle(token)
        return True
    log.error("Error: win32 authentication failed:")
    log.error(" %s", FormatError(GetLastError()))
    return False



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from time import monotonic
from typing import NoReturn
from collections.abc import Sequence

from xpra.util.env import first_time
from xpra.util.str_fn import csv, memoryview_to_bytes
from xpra.codecs.image import ImageWrapper
from xpra.log import Logger

log = Logger("encoding", "argb")


def noswap(image, rgb_formats, _supports_transparency) -> NoReturn:
    pixel_format = image.get_pixel_format()
    raise RuntimeError(f"cannot convert from {pixel_format} to {csv(rgb_formats)} without the argb module")


argb_swap = noswap
try:
    from xpra.codecs.argb import argb
    argb_swap = argb.argb_swap
except ImportError:     # pragma: no cover
    log("no argb_swap", exc_info=True)
    log.warn("Warning: argb swap is missing")


# "pixels_to_bytes" gets patched up by the OSX shadow server
pixels_to_bytes = memoryview_to_bytes


# source format  : [(PIL input format, output format), ..]
PIL_conv: dict[str, Sequence[tuple[str, str]]] = {
    "XRGB": (("XRGB", "RGB"), ),
    # try to drop alpha channel since it isn't used:
    "BGRX": (("BGRX", "RGB"), ("BGRX", "RGBX")),
    # try with alpha first:
    "BGRA": (("BGRA", "RGBA"), ("BGRX", "RGB"), ("BGRX", "RGBX")),
}
# as above but for clients which cannot handle alpha:
PIL_conv_noalpha: dict[str, Sequence[tuple[str, str]]] = {
    "XRGB": (("XRGB", "RGB"), ),
    # try to drop alpha channel since it isn't used:
    "BGRX": (("BGRX", "RGB"), ("BGRX", "RGBX")),
    # try with alpha first:
    "BGRA": (("BGRX", "RGB"), ("BGRA", "RGBA"), ("BGRX", "RGBX")),
}


def rgb_reformat(image: ImageWrapper, rgb_formats, supports_transparency: bool) -> bool:
    """ convert the RGB pixel data into a format supported by the client """
    # need to convert to a supported format!
    pixel_format = image.get_pixel_format()
    pixels = image.get_pixels()
    if not pixels:
        raise RuntimeError(f"failed to get pixels from {image}")
    if pixel_format in ("r210", "BGR565"):
        # try to fall back to argb module
        # (required for r210 which is not handled by PIL directly)
        log("rgb_reformat: using argb_swap for %s", image)
        return argb_swap(image, rgb_formats, supports_transparency)
    modes: Sequence[tuple[str, str]] = ()
    try:
        # pylint: disable=import-outside-toplevel
        from PIL import Image
    except ImportError:
        log("PIL.Image not found!")
        Image = None
    else:
        if supports_transparency:
            modes = PIL_conv.get(pixel_format, ())
        else:
            modes = PIL_conv_noalpha.get(pixel_format, ())
    target_rgb: Sequence[tuple[str, str]] = tuple((im, om) for (im, om) in modes if om in rgb_formats)
    if not target_rgb:
        log("rgb_reformat: no matching target modes for converting %s to %s", image, rgb_formats)
        # try argb module:
        if argb_swap(image, rgb_formats, supports_transparency):
            return True
        warning_key = f"rgb_reformat({pixel_format}, {rgb_formats}, {supports_transparency})"
        if first_time(warning_key):
            log.warn(f"Warning: cannot convert {pixel_format!r} to one of: "+csv(rgb_formats))
        return False
    assert Image
    input_format, target_format = target_rgb[0]
    start = monotonic()
    w = image.get_width()
    h = image.get_height()
    # older versions of PIL cannot use the memoryview directly:
    if isinstance(pixels, memoryview):
        from PIL import __version__ as pil_version
        try:
            major = int(pil_version.split(".")[0])
        except ValueError:
            major = 0
        if major < 10:
            pixels = pixels.tobytes()
    img = Image.frombuffer(target_format, (w, h), pixels, "raw", input_format, image.get_rowstride())
    rowstride = w*len(target_format)    # number of characters is number of bytes per pixel!
    data = img.tobytes("raw", target_format)
    if len(data) != rowstride*h:
        raise RuntimeError(f"expected {rowstride*h} bytes in {target_format} format but got {len(data)}")
    image.set_pixels(data)
    image.set_rowstride(rowstride)
    image.set_pixel_format(target_format)
    end = monotonic()
    log("rgb_reformat(%s, %s, %s) converted from %s (%s bytes) to %s (%s bytes) in %.1fms, rowstride=%s",
        image, rgb_formats, supports_transparency, pixel_format, len(pixels),
        target_format, len(data), (end-start)*1000, rowstride)
    return True



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import shlex
import os.path
from time import monotonic
from subprocess import Popen, PIPE
from threading import Event
from typing import Any
from collections.abc import Callable, Sequence

from xpra.os_util import OSX, POSIX, gi_import
from xpra.util.io import pollwait
from xpra.util.objects import typedict
from xpra.util.str_fn import csv
from xpra.util.env import envbool, osexpand, first_time
from xpra.net.common import PacketType
from xpra.util.thread import start_thread
from xpra.platform.info import get_username
from xpra.platform.paths import get_icon_filename
from xpra.scripts.parsing import audio_option
from xpra.scripts.session import save_session_file
from xpra.server.mixins.stub_server_mixin import StubServerMixin
from xpra.log import Logger

GLib = gi_import("GLib")

log = Logger("server")
audiolog = Logger("audio")

PRIVATE_PULSEAUDIO = envbool("XPRA_PRIVATE_PULSEAUDIO", True)

PA_ENV_WHITELIST = (
    "DBUS_SESSION_BUS_ADDRESS", "DBUS_SESSION_BUS_PID", "DBUS_SESSION_BUS_WINDOWID",
    "DISPLAY", "HOME", "HOSTNAME", "LANG", "PATH",
    "PWD", "SHELL", "XAUTHORITY",
    "XDG_CURRENT_DESKTOP", "XDG_SESSION_TYPE",
    "XPRA_PULSE_SOURCE_DEVICE_NAME", "XPRA_PULSE_SINK_DEVICE_NAME",
)


class AudioServer(StubServerMixin):
    """
    Mixin for servers that handle audio forwarding.
    """
    PREFIX = "audio"

    def __init__(self):
        self.audio_init_done = Event()
        self.audio_init_done.set()
        self.pulseaudio = False
        self.pulseaudio_command = ""
        self.pulseaudio_configure_commands = []
        self.pulseaudio_proc: Popen | None = None
        self.pulseaudio_private_dir = ""
        self.pulseaudio_private_socket = ""
        self.audio_source_plugin = ""
        self.supports_speaker = False
        self.supports_microphone = False
        self.speaker_allowed = False
        self.microphone_allowed = False
        self.speaker_codecs = ()
        self.microphone_codecs = ()
        self.audio_properties = typedict()
        self.av_sync = False

    def init(self, opts) -> None:
        self.audio_source_plugin = opts.audio_source
        self.supports_speaker = audio_option(opts.speaker) in ("on", "off")
        if self.supports_speaker:
            self.speaker_codecs = opts.speaker_codec
        self.supports_microphone = audio_option(opts.microphone) in ("on", "off")
        if self.supports_microphone:
            self.microphone_codecs = opts.microphone_codec
        self.pulseaudio = opts.pulseaudio
        self.pulseaudio_command = opts.pulseaudio_command
        self.pulseaudio_configure_commands = opts.pulseaudio_configure_commands
        log("AudioServer.init(..) supports speaker=%s, microphone=%s",
            self.supports_speaker, self.supports_microphone)
        self.av_sync = opts.av_sync
        log("AudioServer.init(..) av-sync=%s", self.av_sync)

    def setup(self) -> None:
        # the setup code will mostly be waiting for subprocesses to run,
        # so do it in a separate thread
        # and just wait for the results where needed:
        self.audio_init_done.clear()
        start_thread(self.do_audio_setup, "audio-setup", True)
        # we don't use threaded_setup() here because it would delay
        # all the other mixins that use it, for no good reason.

    def do_audio_setup(self) -> None:
        self.init_pulseaudio()
        self.init_audio_options()

    def cleanup(self) -> None:
        self.audio_init_done.wait(5)
        self.cleanup_pulseaudio()

    def get_info(self, _proto) -> dict[str, Any]:
        self.audio_init_done.wait(5)
        info = {}
        if self.pulseaudio is not False:
            info["pulseaudio"] = self.get_pulseaudio_info()
        if self.audio_properties:
            info["audio"] = self.audio_properties
        return {}

    def get_server_features(self, source) -> dict[str, Any]:
        d = {
            "av-sync": {
                "": self.av_sync,
                "enabled": self.av_sync,
            },
        }
        log("get_server_features(%s)=%s", source, d)
        return d

    def init_pulseaudio(self) -> None:
        audiolog("init_pulseaudio() pulseaudio=%s, pulseaudio_command=%s", self.pulseaudio, self.pulseaudio_command)
        if self.pulseaudio is False:
            return
        if not self.pulseaudio_command:
            audiolog.warn("Warning: pulseaudio command is not defined")
            return
        # environment initialization:
        # 1) make sure that the audio subprocess will use the devices
        #    we define in the pulseaudio command
        #    (it is too difficult to parse the pulseaudio_command,
        #    so we just hope that it matches this):
        #    Note: speaker is the source and microphone the sink,
        #    because things are reversed on the server.
        os.environ.update({
            "XPRA_PULSE_SOURCE_DEVICE_NAME": "Xpra-Speaker",
            "XPRA_PULSE_SINK_DEVICE_NAME": "Xpra-Microphone",
        })
        # 2) whitelist the env vars that pulseaudio may use:
        env = {k: v for k, v in self.get_child_env().items() if k in PA_ENV_WHITELIST}
        # 3) use a private pulseaudio server, so each xpra
        #    session can have its own server,
        #    create a directory for each display:
        if PRIVATE_PULSEAUDIO and POSIX and not OSX:
            # pylint: disable=import-outside-toplevel
            from xpra.platform.posix.paths import _get_xpra_runtime_dir, get_runtime_dir
            rd = osexpand(get_runtime_dir())
            if not rd or not os.path.exists(rd) or not os.path.isdir(rd):
                log.warn("Warning: the runtime directory '%s' does not exist,", rd)
                log.warn(" cannot start a private pulseaudio server")
            else:
                xpra_rd = os.environ.get("XPRA_SESSION_DIR", _get_xpra_runtime_dir())
                assert xpra_rd, "bug: no xpra runtime dir"
                display = os.environ.get("DISPLAY", "").lstrip(":")
                if xpra_rd.find(f"/{display}/"):
                    # this is already a per-display directory,
                    # no need to include the display name again:
                    pulse_dirname = "pulse"
                else:
                    pulse_dirname = f"pulse-{display}"
                self.pulseaudio_private_dir = osexpand(os.path.join(xpra_rd, pulse_dirname))
                if not os.path.exists(self.pulseaudio_private_dir):
                    os.mkdir(self.pulseaudio_private_dir, 0o700)
                env["XDG_RUNTIME_DIR"] = self.pulseaudio_private_dir
                self.pulseaudio_private_socket = os.path.join(self.pulseaudio_private_dir, "pulse", "native")
                os.environ["XPRA_PULSE_SERVER"] = self.pulseaudio_private_socket
        cmd = shlex.split(self.pulseaudio_command)
        cmd = list(osexpand(x) for x in cmd)
        # find the absolute path to the command:
        pa_cmd = cmd[0]
        if not os.path.isabs(pa_cmd):
            pa_path = None
            for x in os.environ.get("PATH", "").split(os.path.pathsep):
                t = os.path.join(x, pa_cmd)
                if os.path.exists(t):
                    pa_path = t
                    break
            if not pa_path:
                msg = f"pulseaudio not started: {pa_cmd!r} command not found"
                if self.pulseaudio is None:
                    audiolog.info(msg)
                else:
                    audiolog.warn(msg)
                self.clean_pulseaudio_private_dir()
                return
            cmd[0] = pa_cmd
        started_at = monotonic()

        def pulseaudio_warning() -> None:
            audiolog.warn("Warning: pulseaudio has terminated shortly after startup.")
            audiolog.warn(" pulseaudio is limited to a single instance per user account,")
            audiolog.warn(" and one may be running already for user '%s'.", get_username())
            audiolog.warn(" To avoid this warning, either fix the pulseaudio command line")
            audiolog.warn(" or use the 'pulseaudio=no' option.")

        def pulseaudio_ended(proc) -> None:
            audiolog("pulseaudio_ended(%s) pulseaudio_proc=%s, returncode=%s, closing=%s",
                     proc, self.pulseaudio_proc, proc.returncode, self._closing)
            if self.pulseaudio_proc is None or self._closing:
                # cleared by cleanup already, ignore
                return
            elapsed = monotonic() - started_at
            if elapsed < 2:
                GLib.timeout_add(1000, pulseaudio_warning)
            else:
                audiolog.warn("Warning: the pulseaudio server process has terminated after %i seconds", int(elapsed))
            self.pulseaudio_proc = None

        try:
            audiolog("pulseaudio cmd=%s", " ".join(cmd))
            audiolog("pulseaudio env=%s", env)
            self.pulseaudio_proc = Popen(cmd, env=env)
        except Exception as e:
            audiolog("Popen(%s)", cmd, exc_info=True)
            audiolog.error("Error: failed to start pulseaudio:")
            audiolog.estr(e)
            self.clean_pulseaudio_private_dir()
            return
        self.add_process(self.pulseaudio_proc, "pulseaudio", cmd, ignore=True, callback=pulseaudio_ended)
        if self.pulseaudio_proc:
            save_session_file("pulseaudio.pid", "%s" % self.pulseaudio_proc.pid)
            audiolog.info("pulseaudio server started with pid %s", self.pulseaudio_proc.pid)
            if self.pulseaudio_private_socket:
                audiolog.info(" private server socket path:")
                audiolog.info(" '%s'", self.pulseaudio_private_socket)
                os.environ["PULSE_SERVER"] = "unix:%s" % self.pulseaudio_private_socket

            def configure_pulse() -> None:
                p = self.pulseaudio_proc
                if p is None or p.poll() is not None:
                    return
                for i, x in enumerate(self.pulseaudio_configure_commands):
                    proc = Popen(x, env=env, shell=True)
                    self.add_process(proc, "pulseaudio-configure-command-%i" % i, x, ignore=True)

            GLib.timeout_add(2 * 1000, configure_pulse)

    def cleanup_pulseaudio(self) -> None:
        self.audio_init_done.wait(5)
        proc = self.pulseaudio_proc
        if not proc:
            return
        audiolog("cleanup_pa() process.poll()=%s, pid=%s", proc.poll(), proc.pid)
        if self.is_child_alive(proc):
            self.pulseaudio_proc = None
            audiolog.info("stopping pulseaudio with pid %s", proc.pid)
            try:
                # first we try pactl (required on Ubuntu):
                cmd = ["pactl", "exit"]
                proc = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
                self.add_process(proc, "pactl exit", cmd, True)
                r = pollwait(proc)
                # warning: pactl will return 0 whether it succeeds or not...
                # but we can't kill the process because Ubuntu starts a new one
                if r != 0 and self.is_child_alive(proc):
                    # fallback to using SIGINT:
                    proc.terminate()
            except Exception as e:
                audiolog.warn("cleanup_pulseaudio() error stopping %s", proc, exc_info=True)
                # only log the full stacktrace if the process failed to terminate:
                if self.is_child_alive(proc):
                    audiolog.error("Error: stopping pulseaudio: %s", e, exc_info=True)
            if self.pulseaudio_private_socket and self.is_child_alive(proc):
                # wait for the pulseaudio process to exit,
                # it will delete the socket:
                audiolog("pollwait()=%s", pollwait(proc))
        if self.pulseaudio_private_socket and not self.is_child_alive(proc):
            # wait for the socket to get cleaned up
            # (it should be removed by the pulseaudio server as it exits)
            import time
            now = monotonic()
            while (monotonic() - now) < 1 and os.path.exists(self.pulseaudio_private_socket):
                time.sleep(0.1)
        self.clean_pulseaudio_private_dir()
        if not self.is_child_alive(proc):
            self.do_clean_session_files("pulseaudio.pid")

    def clean_pulseaudio_private_dir(self) -> None:
        if self.pulseaudio_private_dir:
            if os.path.exists(self.pulseaudio_private_socket):
                log.warn("Warning: the pulseaudio private socket file still exists:")
                log.warn(" '%s'", self.pulseaudio_private_socket)
                log.warn(" the private pulseaudio directory containing it will not be removed")
            else:
                import glob
                pulse = os.path.join(self.pulseaudio_private_dir, "pulse")
                native = os.path.join(pulse, "native")
                dirs = []
                dbus_dirs = glob.glob("%s/dbus-*" % self.pulseaudio_private_dir)
                if len(dbus_dirs) == 1:
                    dbus_dir = dbus_dirs[0]
                    if os.path.isdir(dbus_dir):
                        services_dir = os.path.join(dbus_dir, "services")
                        dirs.append(services_dir)
                        dirs.append(dbus_dir)
                dirs += [native, pulse, self.pulseaudio_private_dir]
                path = None
                try:
                    for d in dirs:
                        path = os.path.abspath(d)
                        audiolog("removing private directory '%s'", path)
                        if os.path.exists(path) and os.path.isdir(path):
                            os.rmdir(path)
                    log.info("removing private directory '%s'", self.pulseaudio_private_dir)
                except OSError as e:
                    audiolog("cleanup_pulseaudio() error removing '%s'", path, exc_info=True)
                    audiolog.error("Error: failed to cleanup the pulseaudio private directory")
                    audiolog.error(" '%s'", self.pulseaudio_private_dir)
                    audiolog.estr(e)
                    try:
                        files = os.listdir(path)
                        if files:
                            audiolog.error(f" found %i files in {path!r}:", len(files))
                            for f in files:
                                audiolog.error(f" - {f!r}")
                    except OSError:
                        audiolog.error("cleanup_pulseaudio() error accessing '%s'", path, exc_info=True)

    def init_audio_options(self) -> None:
        def audio_missing(*_args) -> Sequence[str]:
            return ()

        def noaudio() -> None:
            self.supports_speaker = self.supports_microphone = False
            self.speaker_allowed = self.microphone_allowed = False

        parse_codecs: Callable = audio_missing
        if self.supports_speaker or self.supports_microphone:
            try:
                from xpra.audio.common import audio_option_or_all
                parse_codecs = audio_option_or_all
                from xpra.audio.wrapper import query_audio
                self.audio_properties = query_audio()
                if not self.audio_properties:
                    audiolog.info("Audio subsystem query failed, is GStreamer installed?")
                    noaudio()
                    return
                gstv = self.audio_properties.strtupleget("gst.version")
                if gstv:
                    log.info("GStreamer version %s", ".".join(gstv[:3]))
                else:
                    log.info("GStreamer loaded")
            except Exception as e:
                audiolog("failed to query audio", exc_info=True)
                audiolog.error("Error: failed to query audio subsystem:")
                audiolog.estr(e)
                noaudio()
                return
        encoders = self.audio_properties.strtupleget("encoders")
        decoders = self.audio_properties.strtupleget("decoders")
        self.speaker_codecs = parse_codecs("speaker-codec", self.speaker_codecs, encoders)
        self.microphone_codecs = parse_codecs("microphone-codec", self.microphone_codecs, decoders)
        if not self.speaker_codecs:
            self.supports_speaker = False
        if not self.microphone_codecs:
            self.supports_microphone = False
        # query_pulseaudio_properties may access X11,
        # do this from the main thread:
        if bool(self.audio_properties):
            GLib.idle_add(self.query_pulseaudio_properties)
        GLib.idle_add(self.log_audio_properties)
        self.audio_init_done.set()

    def query_pulseaudio_properties(self) -> None:
        try:
            from xpra.audio.pulseaudio.util import set_icon_path, get_info as get_pa_info
            pa_info = get_pa_info()
            audiolog("pulseaudio info=%s", pa_info)
            self.audio_properties.update(pa_info)
            set_icon_path(get_icon_filename("xpra.png"))
        except ImportError as e:
            if POSIX and not OSX:
                log.warn("Warning: failed to set pulseaudio tagging icon:")
                log.warn(" %s", e)

    def log_audio_properties(self) -> None:
        audiolog("init_audio_options speaker: supported=%s, encoders=%s",
                 self.supports_speaker, csv(self.speaker_codecs))
        audiolog("init_audio_options microphone: supported=%s, decoders=%s",
                 self.supports_microphone, csv(self.microphone_codecs))
        audiolog("init_audio_options audio properties=%s", self.audio_properties)

    def get_pulseaudio_info(self) -> dict[str, Any]:
        info = {
            "command": self.pulseaudio_command,
            "configure-commands": self.pulseaudio_configure_commands,
        }
        if self.pulseaudio_proc and self.pulseaudio_proc.poll() is None:
            info["pid"] = self.pulseaudio_proc.pid
        if self.pulseaudio_private_dir and self.pulseaudio_private_socket:
            info["private-directory"] = self.pulseaudio_private_dir
            info["private-socket"] = self.pulseaudio_private_socket
        return info

    def _process_sound_control(self, proto, packet: PacketType) -> None:
        self._process_audio_control(proto, packet)

    def _process_audio_control(self, proto, packet: PacketType) -> None:
        ss = self.get_server_source(proto)
        if not ss:
            return
        audio_control = getattr(ss, "audio_control", None)
        if not audio_control:
            if first_time(f"no-audio-control-{ss}"):
                log.warn(f"Warning: ignoring audio control requests from {ss}")
                log.warn(" audio is not enabled for this connection")
            return
        audio_control(*packet[1:])

    def _process_sound_data(self, proto, packet: PacketType) -> None:
        self._process_sound_data(proto, packet)

    def _process_audio_data(self, proto, packet: PacketType) -> None:
        ss = self.get_server_source(proto)
        if ss:
            ss.audio_data(*packet[1:])

    def init_packet_handlers(self) -> None:
        if self.supports_speaker or self.supports_microphone:
            self.add_packets(f"{AudioServer.PREFIX}-control", main_thread=True)
            self.add_packets(f"{AudioServer.PREFIX}-data")
            self.add_legacy_alias("sound-control", f"{AudioServer.PREFIX}-control")
            self.add_legacy_alias("sound-data", f"{AudioServer.PREFIX}-data")



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import time
import unittest

from xpra.util.objects import AdHocStruct

from unit.test_util import silence_info
from unit.server.mixins.servermixintest_util import ServerMixinTest


class AudioMixinTest(ServerMixinTest):

    @classmethod
    def setUpClass(cls):
        ServerMixinTest.setUpClass()
        from xpra.net import packet_encoding
        packet_encoding.init_all()
        from xpra.net import compression
        compression.init_all()

    def test_audio(self):
        from xpra.server.mixins import audio
        from xpra.server.source.audio import AudioMixin
        from xpra.audio import gstreamer_util
        opts = AdHocStruct()
        opts.audio_source = ""
        opts.speaker = "on"
        opts.speaker_codec = gstreamer_util.CODEC_ORDER
        opts.microphone = "on"
        opts.microphone_codec = ["mp3"]
        opts.audio = True
        opts.pulseaudio = False
        opts.pulseaudio_command = "/bin/true"
        opts.pulseaudio_configure_commands = []
        opts.av_sync = True
        with silence_info(audio, "audiolog"):
            self._test_mixin_class(audio.AudioServer, opts, {
                "audio" : {
                    "receive" : True,
                    "decoders" : gstreamer_util.CODEC_ORDER,
                    },
                }, AudioMixin)
            #init runs in a thread, give it time:
            time.sleep(2)
        if not self.mixin.speaker_codecs:
            print("no speaker codecs available, test skipped")
            return
        codec = self.mixin.speaker_codecs[0]
        with silence_info(gstreamer_util):
            self.handle_packet(("sound-control", "start", codec))
        time.sleep(1)
        self.handle_packet(("sound-control", "fadeout"))
        time.sleep(1)
        self.handle_packet(("sound-control", "stop"))

def main():
    from xpra.os_util import POSIX, OSX
    if POSIX and not OSX:
        #verify that pulseaudio is running:
        #otherwise the tests will fail
        #ie: during rpmbuild
        from subprocess import getstatusoutput
        if getstatusoutput("pactl info")[0]!=0:
            return
        unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.log import Logger

log = Logger("filters")


class WindowPropertyFilter:
    __slots__ = ("property_name", "value", "recurse")

    def __init__(self, property_name, value, recurse=False):
        self.property_name = property_name
        self.value = value
        self.recurse = recurse

    def get_window(self, window):
        return window

    def get_window_value(self, window):
        return window.get_property(self.property_name)

    def matches(self, window):
        w = self.get_window(window)
        log("get_window(%s)=%s", window, w)
        try:
            v = self.get_window_value(w)
            log("%s.matches(%s) %s(..)=%s", self, w, self.get_window_value, v)
        except Exception:
            log("%s.matches(%s) %s(..) error:", self, w, self.get_window_value, exc_info=True)
            return False
        e = self.evaluate(v)
        log("%s.evaluate(%s)=%s (window(%s)=%s)", self, v, e, window, w)
        return e

    def evaluate(self, window_value):
        raise NotImplementedError()


class WindowPropertyIn(WindowPropertyFilter):
    __slots__ = ()

    def evaluate(self, window_value):
        vtypes = set([type(x) for x in self.value])
        if len(vtypes) == 1 and list(vtypes)[0] is str:
            # coerce value to match:
            window_value = str(window_value)
        return window_value in self.value

    def __repr__(self):
        return f"WindowPropertyIn({self.property_name}={self.value}, recurse={self.recurse})"


class WindowPropertyNotIn(WindowPropertyIn):
    __slots__ = ()

    def evaluate(self, window_value):
        return not super().evaluate(window_value)

    def __repr__(self):
        return f"WindowPropertyNotIn({self.property_name}={self.value}, recurse={self.recurse})"


def get_window_filter(object_name, property_name, operator, value):
    oname = object_name.lower()
    if oname not in ("window", "window-parent"):
        raise ValueError("invalid object name '%s'" % object_name)
    recurse = oname == "window-parent"
    if operator == "=":
        window_filter = WindowPropertyIn(property_name, [value], recurse)
    elif operator == "!=":
        window_filter = WindowPropertyNotIn(property_name, [value], recurse)
    else:
        raise ValueError("invalid window filter operator: %s" % operator)
    log("get_window_filter%s=%s", (object_name, property_name, operator, value), window_filter)
    return window_filter



# This file is part of Xpra.
# Copyright (C) 2015 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.log import (
    Logger,
    add_debug_category, add_disabled_category, enable_debug_for, disable_debug_for, get_all_loggers,
    add_backtrace, remove_backtrace,
)
from xpra.net.control.common import ArgsControlCommand
from xpra.util.str_fn import csv

log = Logger("util", "command")


class DebugControl(ArgsControlCommand):
    def __init__(self):
        subcommands = csv(f"'debug {subc}'" for subc in (
            "enable category", "disable category", "status", "mark", "add-backtrace", "remove-backtrace",
        ))
        super().__init__("debug", f"usage: {subcommands}", min_args=1)

    def run(self, *args):
        if not args:
            return "debug subcommands: mark, add-backtrace, remove-backtrace, enable, disable"
        if len(args) == 1 and args[0] == "status":
            return "logging is enabled for: " + csv(str(x) for x in get_all_loggers() if x.is_debug_enabled())
        log_cmd = args[0]
        if log_cmd == "mark":
            for _ in range(10):
                log.info("*" * 80)
            if len(args) > 1:
                log.info("mark: %s", " ".join(args[1:]))
            else:
                log.info("mark")
            for _ in range(10):
                log.info("*" * 80)
            return "mark inserted into logfile"
        if len(args) < 2:
            self.raise_error("not enough arguments")
        if log_cmd == "add-backtrace":
            expressions = args[1:]
            add_backtrace(*expressions)
            return f"added backtrace expressions {expressions}"
        if log_cmd == "remove-backtrace":
            expressions = args[1:]
            remove_backtrace(*expressions)
            return f"removed backtrace expressions {expressions}"
        if log_cmd not in ("enable", "disable"):
            self.raise_error("only 'enable' and 'disable' verbs are supported")
        # each argument is a group
        loggers = []
        groups = args[1:]
        for group in groups:
            # and each group is a list of categories
            # preferably separated by "+",
            # but we support "," for backwards compatibility:
            categories = [v.strip() for v in group.replace("+", ",").split(",")]
            if log_cmd == "enable":
                add_debug_category(*categories)
                loggers += enable_debug_for(*categories)
            elif log_cmd == "disable":
                add_disabled_category(*categories)
                loggers += disable_debug_for(*categories)
            else:
                raise ValueError(f"invalid log command {log_cmd}")
        if not loggers:
            log.info("%s debugging, no new loggers matching: %s", log_cmd, csv(groups))
        else:
            log.info("%sd debugging for:", log_cmd)
            for logger in loggers:
                log.info(" - %s", logger)
        return f"logging {log_cmd}d for " + (csv(loggers) or "<no match found>")



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

from xpra.util.objects import AdHocStruct
from xpra.os_util import POSIX
from unit.test_util import silence_info
from unit.server.mixins.servermixintest_util import ServerMixinTest


class FilePrintMixinTest(ServerMixinTest):

    def create_test_sockets(self):
        if not POSIX:
            return ()
        return [
            #socktype, socket, sockpath, cleanup_socket
            ("socket", None, "/fake/path", None)
            ]

    def test_fileprint(self):
        from xpra.server.mixins import fileprint
        opts = AdHocStruct()
        opts.file_transfer = "yes"
        opts.file_size_limit = 10
        opts.printing = "yes"
        opts.open_files = "no"
        opts.open_url = "yes"
        opts.open_command = ""
        opts.lpadmin = "/usr/sbin/lpadmin"
        opts.lpinfo = "/usr/sbin/lpinfo"
        opts.add_printer_options = ""
        opts.postscript_printer = ""
        opts.pdf_printer = ""
        with silence_info(fileprint, "printlog"):
            self._test_mixin_class(fileprint.FilePrintServer, opts)

def main():
    unittest.main()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2025 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from io import BytesIO
from tkinter import Toplevel, Label
from PIL import Image, ImageTk

from xpra.util.objects import typedict
from xpra.log import Logger

log = Logger("client", "window")


def modifiers(value: int) -> list[str]:
    mod_mask = {
        1: "shift",
        2: "Caps_Lock",
        4: "control",
        8: "Alt_L",
        0x10: "Num_Lock",
        0x80: "Alt_R",
    }
    mods = []
    for mask, mod in mod_mask.items():
        if value & mask:
            mods.append(mod)
    return mods


class ClientWindow(Toplevel):
    def __init__(self, client, wid: int, x: int, y: int, w: int, h: int, metadata: dict):
        from xpra.client.tk.client import app
        super().__init__(app)
        self.withdraw()
        self.client = client
        self.send = client.send
        self.wid = wid
        self.metadata = typedict(metadata)
        self.overrideredirect(self.metadata.boolget("override-redirect", False))
        self.title(self.metadata.strget("title", ""))
        self.geometry(f"{w}x{h}+{x}+{y}")
        self.update_metadata(self.metadata)
        self.update_idletasks()

        self.backing = Image.new(mode="RGBA", size=(w, h))
        self.photo_image = ImageTk.PhotoImage(self.backing)
        self.label = Label(self, image=self.photo_image)
        self.label.pack()

        self.bind("<Motion>", self.on_mouse_motion)
        for button in range(1, 3):
            self.bind(f"<Button-{button}>", self.on_mouse_button_press)
            self.bind(f"<ButtonRelease-{button}>", self.on_mouse_button_release)
        # self.bind("<MouseWheel>", self.on_mouse_wheel)
        self.bind("<FocusIn>", self.on_focus_in)
        self.bind("<FocusOut>", self.on_focus_out)
        self.bind("<Key>", self.on_key_press)
        self.bind("<KeyRelease>", self.on_key_release)
        self.bind("<Configure>", self.on_configure)
        self.bind("<Map>", self.on_map)
        self.bind("<Unmap>", self.on_unmap)
        self.bind("<Destroy>", self.on_destroy)
        self.protocol("WM_DELETE_WINDOW", self.on_destroy)

    def update_metadata(self, metadata: typedict):
        self.metadata.update(metadata)
        if "title" in metadata:
            self.title(metadata.strget("title"))
        else:
            log(f"unused window metadata {self.wid}: {metadata}")

    def show(self) -> None:
        self.deiconify()

    def on_focus_in(self, event) -> None:
        log(f"focus-in: {event!r}")
        self.client.update_focus(self.wid)

    def on_focus_out(self, event) -> None:
        pass

    def on_hide(self) -> None:
        self.client.send("unmap-window", self.wid)

    def on_key_press(self, event) -> None:
        log(f"key press: {event!r}")
        self.on_key(True, event)

    def on_key_release(self, event) -> None:
        log(f"key release: {event!r}")
        self.on_key(False, event)

    def on_key(self, pressed: bool, event) -> None:
        mods = modifiers(event.state)
        keyval = 0
        keyname = event.keysym
        string = event.char
        keycode = event.keycode
        group = 0
        self.client.send("key-action", self.wid, keyname, pressed, mods, keyval, string, keycode, group)

    def on_mouse_motion(self, event) -> None:
        log(f"mouse motion: {event!r}")
        device_id = -1
        seq = 0
        pos = (event.x, event.y)
        self.client.send("pointer", device_id, seq, self.wid, pos, {})

    def on_mouse_button_press(self, event) -> None:
        log(f"mouse button press: {event!r}")
        self.on_mouse_button_event(True, event)

    def on_mouse_button_release(self, event) -> None:
        log(f"mouse button release: {event!r}")
        self.on_mouse_button_event(False, event)

    def on_mouse_button_event(self, pressed: bool, event) -> None:
        seq = 0
        button = event.num
        pos = (event.x, event.y)
        self.client.send("pointer-button", -1, seq, self.wid, button, pressed, pos, {})

    def on_map(self, event) -> None:
        log(f"map: {event!r}")
        x = self.winfo_x()
        y = self.winfo_y()
        w = self.winfo_width()
        h = self.winfo_height()
        self.client.send("map-window", self.wid, x, y, w, h, {}, ())

    def on_unmap(self, event) -> None:
        self.client.send("close-window", self.wid)

    def on_configure(self, event) -> None:
        log(f"configure: {event}")
        counter = 0
        props = {}
        x, y, w, h = event.x, event.y, event.width, event.height
        state = ()
        skip_geometry = False
        self.client.send("configure-window", self.wid, x, y, w, h, props, counter, state, skip_geometry)

    def on_destroy(self, event=None) -> None:
        self.client.send("close-window", self.wid)

    def on_resize(self, event) -> None:
        counter = 0
        props = {}
        x = self.winfo_x()
        y = self.winfo_y()
        w = self.winfo_width()
        h = self.winfo_height()
        state = ()
        skip_geometry = False
        self.client.send("configure-window", self.wid, x, y, w, h, props, counter, state, skip_geometry)

    def draw(self, x, y, w, h, coding, data, stride) -> None:
        if coding not in ("png", "jpg", "webp"):
            raise ValueError(f"unsupported format {coding!r}")
        img = Image.open(BytesIO(data))
        self.backing.paste(img, (x, y))
        self.photo_image.paste(self.backing)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from collections.abc import Sequence

from xpra.util.str_fn import csv
from xpra.util.parsing import parse_str_dict
from xpra.os_util import getuid, getgid
from xpra.auth.common import SessionData
from xpra.auth.sys_auth_base import SysAuthenticator, log


class SQLAuthenticator(SysAuthenticator):
    CLIENT_USERNAME = True

    def __init__(self, **kwargs):
        self.password_query: str = kwargs.pop("password_query", "SELECT password FROM users WHERE username=(%s)")
        self.sessions_query: str = kwargs.pop("sessions_query",
                                              "SELECT uid, gid, displays, env_options, session_options FROM users WHERE username=(%s) AND password=(%s)")  # noqa: E501
        super().__init__(**kwargs)
        self.authenticate_check = self.authenticate_hmac

    def db_cursor(self, *sqlargs):
        raise NotImplementedError()

    def get_passwords(self) -> Sequence[str]:
        cursor = self.db_cursor(self.password_query, (self.username,))
        data = cursor.fetchall()
        if not data:
            log.info("username {self.username!r} was not found in SQL authentication database")
            return ()
        return tuple(str(x[0]) for x in data)

    def get_sessions(self) -> SessionData | None:
        cursor = self.db_cursor(self.sessions_query, (self.username, self.password_used or ""))
        data = cursor.fetchone()
        if not data:
            return None
        return self.parse_session_data(data)

    def parse_session_data(self, data) -> SessionData | None:
        displays: list[str] = []
        env_options: dict[str, str] = {}
        session_options: dict[str, str] = {}
        try:
            uid = int(data[0] or "0")
            gid = int(data[1] or "0")
            if len(data) > 2:
                displays = [x.strip() for x in str(data[2]).split(",")]
            if len(data) > 3:
                env_options = parse_str_dict(str(data[3]), ";")
            if len(data) > 4:
                session_options = parse_str_dict(str(data[4]), ";")
        except Exception as e:
            log("parse_session_data() error on row %s", data, exc_info=True)
            log.error("Error: sqlauth database row parsing problem:")
            log.estr(e)
            uid = self.get_uid()
            gid = self.get_gid()
        return uid, gid, displays, env_options, session_options


class DatabaseUtilBase:

    def __init__(self, uri: str):
        self.uri = uri
        self.param = "?"

    def exec_database_sql_script(self, cursor_cb, *sqlargs):
        raise NotImplementedError()

    def create(self) -> None:
        sql = ("CREATE TABLE users ("
               "username VARCHAR(255) NOT NULL, "
               "password VARCHAR(255), "
               "uid VARCHAR(63), "
               "gid VARCHAR(63), "
               "displays VARCHAR(8191), "
               "env_options VARCHAR(8191), "
               "session_options VARCHAR(8191))")
        self.exec_database_sql_script(None, sql)

    def add_user(self, username: str, password: str, uid: int = getuid(), gid: int = getgid(),
                 displays="", env_options="", session_options="") -> None:
        sql = "INSERT INTO users(username, password, uid, gid, displays, env_options, session_options) " \
              "VALUES(%s, %s, %s, %s, %s, %s, %s)" % ((self.param,) * 7)
        self.exec_database_sql_script(None, sql,
                                      (username, password, uid, gid, displays, env_options, session_options))

    def remove_user(self, username: str, password: str = "") -> None:
        sql = "DELETE FROM users WHERE username=%s" % self.param
        sqlargs: Sequence[str] = (username,)
        if password:
            sql += " AND password=%s" % self.param
            sqlargs = (username, password)
        self.exec_database_sql_script(None, sql, sqlargs)

    def list_users(self) -> None:
        fields = ("username", "password", "uid", "gid", "displays", "env_options", "session_options")

        def fmt(values, sizes):
            s = ""
            for i, field in enumerate(values):
                if i == 0:
                    s += "|"
                s += ("%s" % field).rjust(sizes[i]) + "|"
            return s

        def cursor_callback(cursor):
            rows = cursor.fetchall()
            if not rows:
                print("no rows found")
                cursor.close()
                return
            print("%i rows found:" % len(rows))
            # calculate max size for each field:
            sizes = [len(x) + 1 for x in fields]
            for row in rows:
                for i, value in enumerate(row):
                    sizes[i] = max(sizes[i], len(str(value)) + 1)
            total = sum(sizes) + len(fields) + 1
            print("-" * total)
            print(fmt((field.replace("_", " ") for field in fields), sizes))
            print("-" * total)
            for row in rows:
                print(fmt(row, sizes))
            cursor.close()

        sql = "SELECT %s FROM users" % csv(fields)
        self.exec_database_sql_script(cursor_callback, sql)

    def authenticate(self, username: str, password: str) -> None:
        auth_class = self.get_authenticator_class()
        a = auth_class(username, self.uri)
        passwords = a.get_passwords()
        assert passwords
        log("authenticate: got %i passwords", len(passwords))
        assert password in passwords
        a.password_used = password
        sessions = a.get_sessions()
        assert sessions, "no sessions found"
        log("sql authentication success, found sessions: %s", sessions)

    def get_authenticator_class(self) -> type:
        raise NotImplementedError()


def run_dbutil(database_util_class=DatabaseUtilBase, conn_str="databaseURI", argv=()) -> int:
    def usage(msg="invalid number of arguments"):
        print(msg)
        print("usage:")
        print(" %s %s create" % (argv[0], conn_str))
        print(" %s %s list" % (argv[0], conn_str))
        print(" %s %s add username password [uid, gid, displays, env_options, session_options]" % (argv[0], conn_str))
        print(" %s %s remove username [password]" % (argv[0], conn_str))
        print(" %s %s authenticate username password" % (argv[0], conn_str))
        return 1

    # pylint: disable=import-outside-toplevel
    from xpra.platform import program_context
    with program_context("SQL Auth", "SQL Auth"):
        nargs = len(argv)
        if nargs < 3:
            return usage()
        uri = argv[1]
        dbutil = database_util_class(uri)
        cmd = argv[2]
        if cmd == "create":
            if nargs != 3:
                return usage()
            dbutil.create()
        elif cmd == "add":
            if nargs < 5 or nargs > 10:
                return usage()
            dbutil.add_user(*argv[3:])
        elif cmd == "remove":
            if nargs not in (4, 5):
                return usage()
            dbutil.remove_user(*argv[3:])
        elif cmd == "list":
            if nargs != 3:
                return usage()
            dbutil.list_users()
        elif cmd == "authenticate":
            if nargs != 5:
                return usage()
            dbutil.authenticate(*argv[3:])
        else:
            return usage("invalid command '%s'" % cmd)
    return 0



# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from typing import Any
from collections.abc import Callable

from xpra.util.screen import prettify_plug_name
from xpra.util.str_fn import csv
from xpra.util.parsing import parse_simple_dict
from xpra.util.env import envbool, SilenceWarningsContext
from xpra.common import XPRA_APP_ID, noop
from xpra.os_util import POSIX, OSX, gi_import
from xpra.scripts.config import str_to_bool
from xpra.server import features
from xpra.server.shadow.root_window_model import RootWindowModel
from xpra.server.gtk_server import GTKServerBase
from xpra.server.shadow.shadow_server_base import ShadowServerBase
from xpra.codecs.constants import TransientCodecException, CodecStateException
from xpra.gtk.util import get_default_root_window
from xpra.gtk.info import get_screen_sizes
from xpra.gtk.pixbuf import get_icon_pixbuf
from xpra.net.compression import Compressed
from xpra.log import Logger

GLib = gi_import("GLib")
Gtk = gi_import("Gtk")
Gdk = gi_import("Gdk")

traylog = Logger("tray")
notifylog = Logger("notify")
screenlog = Logger("screen")
log = Logger("shadow")

MULTI_WINDOW = envbool("XPRA_SHADOW_MULTI_WINDOW", True)


def parse_geometry(s) -> list[int]:
    try:
        parts = s.split("@")
        if len(parts) == 1:
            x = y = 0
        else:
            x, y = (int(v.strip(" ")) for v in parts[1].split("x"))
        w, h = (int(v.strip(" ")) for v in parts[0].split("x"))
        geometry = [x, y, w, h]
        screenlog("capture geometry: %s", geometry)
        return geometry
    except ValueError:
        screenlog("failed to parse geometry %r", s, exc_info=True)
        screenlog.error("Error: invalid display geometry specified: %r", s)
        screenlog.error(" use the format: WIDTHxHEIGHT@x,y")
        raise


def parse_geometries(s) -> list:
    g = []
    for geometry_str in s.split("/"):
        if geometry_str:
            g.append(parse_geometry(geometry_str))
    return g


def checkitem(title: str, cb: Callable[[Any], None] = noop, active=False) -> Gtk.CheckMenuItem:
    check_item = Gtk.CheckMenuItem(label=title)
    check_item.set_active(active)
    if cb:
        check_item.connect("toggled", cb)
    check_item.show()
    return check_item


def get_icon_image(icon_name: str):
    from xpra.platform.gui import get_icon_size
    size = get_icon_size()
    from xpra.gtk.widget import scaled_image
    with log.trap_error(f"Error loading image from icon {icon_name!r} with size {size}"):
        pixbuf = get_icon_pixbuf(icon_name)
        traylog("get_image(%s, %s) pixbuf=%s", icon_name, size, pixbuf)
        if not pixbuf:
            return None
        return scaled_image(pixbuf, size)


class GTKShadowServerBase(ShadowServerBase, GTKServerBase):

    def __init__(self, attrs: dict[str, str]):
        ShadowServerBase.__init__(self, get_default_root_window())
        GTKServerBase.__init__(self)
        self.session_type = "shadow"
        self.multi_window = str_to_bool(attrs.get("multi-window", True))
        # for managing the systray
        self.tray_menu = None
        self.tray_menu_shown = False
        self.tray_widget = None
        self.tray = False
        self.tray_icon = None

    def init(self, opts) -> None:
        GTKServerBase.init(self, opts)
        ShadowServerBase.init(self, opts)
        self.tray = opts.tray
        self.tray_icon = opts.tray_icon
        if self.tray:
            self.setup_tray()

    def cleanup(self) -> None:
        self.cleanup_tray()
        ShadowServerBase.cleanup(self)
        GTKServerBase.cleanup(self)  # @UndefinedVariable

    def client_startup_complete(self, ss) -> None:
        GTKServerBase.client_startup_complete(self, ss)
        if not self.tray_icon:
            self.set_tray_icon("server-connected")

    def last_client_exited(self) -> None:
        log("last_client_exited() mapped=%s", self.mapped)
        for wid in tuple(self.mapped):
            self.stop_refresh(wid)
        # revert to default icon:
        if not self.tray_icon:
            self.set_tray_icon("server-notconnected")
        GTKServerBase.last_client_exited(self)

    def make_hello(self, source) -> dict[str, Any]:
        caps = ShadowServerBase.make_hello(self, source)
        caps.update(GTKServerBase.make_hello(self, source))
        if "features" in source.wants:
            caps["screen_sizes"] = get_screen_sizes()
        return caps

    def get_info(self, proto=None, *args) -> dict[str, Any]:
        info = ShadowServerBase.get_info(self, proto, *args)
        info.update(GTKServerBase.get_info(self, proto, *args))
        return info

    def accept_client_ssh_agent(self, uuid: str, ssh_auth_sock: str) -> None:
        log("accept_client_ssh_agent: not setting up ssh agent forwarding for shadow servers")

    def refresh(self) -> bool:
        log("refresh() mapped=%s, capture=%s", self.mapped, self.capture)
        if not self.mapped:
            self.refresh_timer = 0
            return False
        self.refresh_window_models()
        if self.capture:
            try:
                if not self.capture.refresh():
                    # capture doesn't have any screen updates,
                    # so we can skip calling damage
                    # (this shortcut is only used with nvfbc)
                    return True
            except TransientCodecException as tce:
                log("refresh()", exc_info=True)
                log.warn("Warning: transient codec exception:")
                log.warn(" %s", tce)
                self.recreate_window_models()
                return False
            except CodecStateException as cse:
                log("refresh()", exc_info=True)
                log.warn("Warning: codec state exception:")
                log.warn(" %s", cse)
                self.recreate_window_models()
                return False
        self.refresh_windows()
        return True

    def refresh_windows(self) -> None:
        for window in self._id_to_window.values():
            self.refresh_window(window)

    ############################################################################
    # handle monitor changes

    def send_updated_screen_size(self) -> None:
        log("send_updated_screen_size")
        super().send_updated_screen_size()
        if features.windows:
            self.recreate_window_models()

    def recreate_window_models(self) -> None:
        # remove all existing models and re-create them:
        for model in tuple(self._window_to_id.keys()):
            self._remove_window(model)
        self.cleanup_capture()
        for model in self.makeRootWindowModels():
            self._add_new_window(model)

    def setup_capture(self):
        raise NotImplementedError()

    def get_root_window_model_class(self) -> type:
        return RootWindowModel

    def get_shadow_monitors(self) -> list[tuple[str, int, int, int, int, int]]:
        display = self.root.get_display()
        if not display:
            return []
        n = display.get_n_monitors()
        monitors = []
        for i in range(n):
            m = display.get_monitor(i)
            geom = m.get_geometry()
            try:
                scale_factor = m.get_scale_factor()
            except Exception as e:
                screenlog("no scale factor: %s", e)
                scale_factor = 1
            else:
                screenlog("scale factor for monitor %i: %i", i, scale_factor)
            plug_name = m.get_model()
            monitors.append((plug_name, geom.x, geom.y, geom.width, geom.height, scale_factor))
        screenlog("get_shadow_monitors()=%s", monitors)
        return monitors

    def makeRootWindowModels(self) -> list:
        screenlog("makeRootWindowModels() root=%s, display_options=%s", self.root, self.display_options)
        self.capture = self.setup_capture()
        if not self.capture:
            raise RuntimeError("failed to instantiate a capture backend")
        log.info(f"capture using {self.capture.get_type()}")
        model_class = self.get_root_window_model_class()
        models = []
        display_name = prettify_plug_name(self.root.get_screen().get_display().get_name())
        monitors = self.get_shadow_monitors()
        match_str = None
        geometries = None
        if "=" in self.display_options:
            # parse the display options as a dictionary:
            opt_dict = parse_simple_dict(self.display_options)
            windows = opt_dict.get("windows")
            if windows:
                self.window_matches = windows.split("/")
                return self.makeDynamicWindowModels()
            match_str = opt_dict.get("plug")
            self.multi_window = str_to_bool(opt_dict.get("multi-window", self.multi_window))
            geometries_str = opt_dict.get("geometry")
            if geometries_str:
                geometries = parse_geometries(geometries_str)
        else:
            try:
                geometries = parse_geometries(self.display_options)
            except ValueError:
                match_str = self.display_options
        log(f"makeRootWindowModels() multi_window={self.multi_window}")
        if not self.multi_window or geometries:
            for geometry in (geometries or (self.root.get_geometry()[:4],)):
                model = model_class(self.root, self.capture, display_name, geometry)
                models.append(model)
            return models
        found = []
        screenlog("capture inputs matching %r", match_str or "all")
        for i, monitor in enumerate(monitors):
            plug_name, x, y, width, height, scale_factor = monitor
            title = display_name
            if plug_name or i > 1:
                title = plug_name or str(i)
            found.append(plug_name or title)
            if match_str and not (title in match_str or plug_name in match_str):
                screenlog.info(" skipped monitor %s", plug_name or title)
                continue
            geometry = (x, y, width, height)
            model = model_class(self.root, self.capture, title, geometry)
            models.append(model)
            screenlog("monitor %i: %10s geometry=%s, scale factor=%s", i, title, geometry, scale_factor)
        screenlog("makeRootWindowModels()=%s", models)
        if not models and match_str:
            screenlog.warn("Warning: no monitors found matching %r", match_str)
            screenlog.warn(" only found: %s", csv(found))
        return models

    def makeDynamicWindowModels(self):
        assert self.window_matches
        raise NotImplementedError("dynamic window shadow is not implemented on this platform")

    def refresh_window_models(self) -> None:
        if not self.window_matches or not features.windows:
            return
        # update the window models which may have changed,
        # some may have disappeared, new ones created,
        # or they may just have changed their geometry:
        try:
            windows = self.makeDynamicWindowModels()
        except Exception as e:
            log("refresh_window_models()", exc_info=True)
            log.error("Error refreshing window models")
            log.estr(e)
            return
        # build a map of window identifier -> window model:
        xid_to_window = {window.get_id(): window for window in windows}
        log("xid_to_window(%s)=%s", windows, xid_to_window)
        sources = self.window_sources()
        for wid, window in tuple(self._id_to_window.items()):
            xid = window.get_id()
            new_model = xid_to_window.pop(xid, None)
            if new_model is None:
                # window no longer exists:
                self._remove_window(window)
                continue
            resized = window.geometry[2:] != new_model.geometry[2:]
            window.geometry = new_model.geometry
            if resized:
                # it has been resized:
                window.geometry = new_model.geometry
                window.notify("size-hints")
                for ss in sources:
                    ss.resize_window(wid, window, window.geometry[2], window.geometry[3])
        # any models left are new windows:
        for window in xid_to_window.values():
            self._add_new_window(window)
            self.refresh_window(window)

    def _adjust_pointer(self, proto, device_id, wid: int, opointer) -> list[int] | None:
        window = self._id_to_window.get(wid)
        if wid > 0 and not window:
            self.suspend_cursor(proto)
            return None
        pointer = super()._adjust_pointer(proto, device_id, wid, opointer)
        ax = x = int(pointer[0])
        ay = y = int(pointer[1])
        if window:
            # the window may be at an offset (multi-window for multi-monitor):
            wx, wy, ww, wh = window.get_geometry()
            # or maybe the pointer is off-screen:
            if x < 0 or x >= ww or y < 0 or y >= wh:
                self.suspend_cursor(proto)
                return None
            # note: with x11 shadow servers,
            # X11ServerCore._get_pointer_abs_coordinates() will recalculate
            # the absolute coordinates from the relative ones,
            # and it should end up with the same values we calculated here
            ax = x + wx
            ay = y + wy
        self.restore_cursor(proto)
        return [ax, ay] + list(pointer[2:])

    def get_pointer_position(self):
        return self.root.get_pointer()[-3:-1]

    def get_notification_tray(self):
        return self.tray_widget

    def get_notifier_classes(self) -> list[type]:
        ncs = list(ShadowServerBase.get_notifier_classes(self))
        try:
            from xpra.gtk.notifier import GTKNotifier  # pylint: disable=import-outside-toplevel
            ncs.append(GTKNotifier)
        except Exception as e:
            notifylog("get_notifier_classes()", exc_info=True)
            notifylog.warn("Warning: cannot load GTK notifier:")
            notifylog.warn(" %s", e)
        return ncs

    ############################################################################
    # system tray methods, mostly copied from the gtk client...
    # (most of these should probably be moved to a common location instead)

    def cleanup_tray(self) -> None:
        tw = self.tray_widget
        traylog("cleanup_tray() tray_widget=%s", tw)
        if tw:
            self.tray_widget = None
            tw.cleanup()

    def setup_tray(self) -> None:
        if OSX:
            return
        display = Gdk.Display.get_default()
        if not display:
            # usually this is wayland shadow server:
            traylog("no access to the display, cannot setup tray")
            return
        try:
            # menu:
            label = "Xpra Shadow Server"
            display = os.environ.get("DISPLAY")
            if POSIX and display:
                label = f"Xpra {display} Shadow Server"
            self.tray_menu = Gtk.Menu()
            with SilenceWarningsContext(DeprecationWarning):
                self.tray_menu.set_title(label)
            title_item = Gtk.MenuItem()
            title_item.set_label(label)
            title_item.set_sensitive(False)
            title_item.show()
            self.tray_menu.append(title_item)

            def show_about(*_args):
                from xpra.gtk.dialogs.about import about  # pylint: disable=import-outside-toplevel
                about()

            self.tray_menu.append(self.traymenuitem("About Xpra", "information.png", cb=show_about))
            if features.windows:
                def readonly_toggled(menuitem) -> None:
                    log("readonly_toggled(%s)", menuitem)
                    ro = menuitem.get_active()
                    if ro != self.readonly:
                        self.readonly = ro
                        self.setting_changed("readonly", ro)

                self.tray_menu.append(checkitem("Read-only", cb=readonly_toggled, active=self.readonly))
            self.tray_menu.append(self.traymenuitem("Exit", "quit.png", cb=self.tray_exit_callback))
            self.tray_menu.append(self.traymenuitem("Close Menu", "close.png", cb=self.close_tray_menu))
            # maybe add: session info, clipboard, sharing, etc
            # control: disconnect clients
            self.tray_menu.connect("deactivate", self.tray_menu_deactivated)
            self.tray_widget = self.make_tray_widget()
            self.set_tray_icon(self.tray_icon or "server-notconnected")
        except ImportError as e:
            traylog("setup_tray()", exc_info=True)
            traylog.warn("Warning: failed to load systemtray:")
            traylog.warn(" %s", e)
        except Exception as e:
            traylog("error setting up %s", self.tray_widget, exc_info=True)
            traylog.error("Error setting up system tray:")
            traylog.estr(e)

    def make_tray_widget(self):
        # pylint: disable=import-outside-toplevel
        from xpra.platform.gui import get_native_system_tray_classes
        classes = get_native_system_tray_classes()
        try:
            from xpra.client.gtk3.statusicon_tray import GTKStatusIconTray
            classes.append(GTKStatusIconTray)
        except ImportError:
            traylog("no GTKStatusIconTray", exc_info=True)
        traylog("tray classes: %s", classes)
        if not classes:
            traylog.error("Error: no system tray implementation available")
            return None
        errs = []
        for c in classes:
            try:
                w = c(self, XPRA_APP_ID, self.tray_menu, "Xpra Shadow Server",
                      icon_filename="server-notconnected.png",
                      click_cb=self.tray_click_callback, exit_cb=self.tray_exit_callback)
                if w:
                    traylog(f"server system tray widget using {c}(..)={w}")
                    return w
                traylog(f"{c}(..) returned None")
                errs.append((c, "returned None"))
            except Exception as e:
                traylog(f"{c}(..)", exc_info=True)
                errs.append((c, e))
        traylog.error("Error: all system tray implementations have failed")
        for c, err in errs:
            traylog.error(" %s: %s", c, err)
        return None

    def set_tray_icon(self, filename: str) -> None:
        if not self.tray_widget:
            return
        try:
            self.tray_widget.set_icon(filename)
        except Exception as e:
            traylog.warn("Warning: failed to set tray icon to %s", filename)
            traylog.warn(" %s", e)

    def traymenuitem(self, title: str, icon_name="", tooltip="", cb: Callable = noop) -> Gtk.ImageMenuItem:
        """ Utility method for easily creating an ImageMenuItem """
        # pylint: disable=import-outside-toplevel
        from xpra.gtk.widget import menuitem
        image = None
        if icon_name:
            image = get_icon_image(icon_name)
        return menuitem(title, image, tooltip, cb)

    def tray_menu_deactivated(self, *_args) -> None:
        self.tray_menu_shown = False

    def tray_click_callback(self, button: int, pressed: int, time=0) -> None:
        traylog("tray_click_callback(%s, %s, %i) tray menu=%s, shown=%s",
                button, pressed, time, self.tray_menu, self.tray_menu_shown)
        if pressed:
            self.close_tray_menu()
        else:
            # status icon can give us a position function:
            # except this doesn't work and nothing happens!
            # position_menu = self.tray_widget.tray_widget.position_menu
            # pos = position_menu(self.tray_menu, x, y, self.tray_widget.tray_widget)
            if POSIX and not OSX:
                self.tray_menu.popup_at_pointer()
            else:
                with SilenceWarningsContext(DeprecationWarning):
                    self.tray_menu.popup(None, None, None, None, button, time)
            self.tray_menu_shown = True

    def tray_exit_callback(self, *_args) -> None:
        self.close_tray_menu()
        GLib.idle_add(self.clean_quit, False)

    def close_tray_menu(self, *_args) -> None:
        if self.tray_menu_shown:
            self.tray_menu.popdown()
            self.tray_menu_shown = False

    ############################################################################
    # screenshot
    def do_make_screenshot_packet(self):
        assert len(self._id_to_window) == 1, "multi root window screenshot not implemented yet"
        rwm = self._id_to_window.values()[0]
        w, h, encoding, rowstride, data = rwm.take_screenshot()
        assert encoding == "png"  # use fixed encoding for now
        return ["screenshot", w, h, encoding, rowstride, Compressed(encoding, data)]



# -*- coding: utf-8 -*-
# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# this used to be implemented using a gtk.gdk.Rectangle
# but we don't want its union() behaviour which can be too expensive

#cython: boundscheck=False, wraparound=False, overflowcheck=False

cdef inline int MIN(int a, int b):   # pylint: disable=syntax-error
    if a<=b:
        return a
    return b

cdef inline int MAX(int a, int b):
    if a>=b:
        return a
    return b


cdef class rectangle:

    cdef readonly int x, y, width, height
    cdef readonly long hash

    def __init__(self, const int x, const int y, const int w, const int h):
        if w<0 or h<0:
            raise ValueError(f"invalid dimensions: {w}x{h}")
        self.x = x
        self.y = y
        self.width = w
        self.height = h
        self.hash = (self.x&0xffff)<<48+(self.y&0xffff)<<32+(self.width&0xffff)<<16+(self.height&0xffff)

    def __hash__(self):
        return self.hash

    def __str__(self):
        return "rectangle(%i, %i, %i, %i)" % (self.x, self.y, self.width, self.height)

    def __repr__(self):
        return "R(%i, %i, %i, %i)" % (self.x, self.y, self.width, self.height)

    def __richcmp__(self, object other, const int op):
        if not isinstance(other, rectangle):
            raise ValueError("cannot compare %s and %s" % (type(self), type(other)))
        cdef rectangle o = other
        if op==2:   #==
            return self.hash==o.hash and self.x==o.x and self.y==o.y and self.width==o.width and self.height==o.height
        elif op==3: #!=
            return self.x!=o.x or self.y!=o.y or self.width!=o.width or self.height!=o.height
        elif op==0: #<
            return self.x<o.x or self.y<o.y or self.width<o.width or self.height<o.height
        elif op==1: #<=
            return self.x<=o.x or self.y<=o.y or self.width<=o.width or self.height<=o.height
        elif op==4: #>
            return self.x>o.x or self.y>o.y or self.width>o.width or self.height>o.height
        elif op==5: #>=
            return self.x>=o.x or self.y>=o.y or self.width>=o.width or self.height>=o.height
        else:
            raise ValueError("invalid richcmp operator: %s" % op)

    def intersects(self, const int x, const int y, const int w, const int h) -> bool:
        cdef int  ix = MAX(self.x, x)
        cdef int  iw = MIN(self.x+self.width, x+w) - ix
        if iw<=0:
            return False
        cdef int  iy = MAX(self.y, y)
        cdef int  ih = MIN(self.y+self.height, y+h) - iy
        return ih>0

    def intersects_rect(self, rectangle rect) -> bool:
        return self.intersects(rect.x, rect.y, rect.width, rect.height)

    def intersection(self, const int x, const int y, const int w, const int h):
        """ returns the rectangle containing the intersection with the given area,
            or None
        """
        cdef int ix = MAX(self.x, x)
        cdef int iw = MIN(self.x+self.width, x+w) - ix
        if iw<=0:
            return None
        cdef int iy = MAX(self.y, y)
        cdef int ih = MIN(self.y+self.height, y+h) - iy
        if ih<=0:
            return None
        return rectangle(ix, iy, iw, ih)

    def intersection_rect(self, rectangle rect):
        return self.intersection(rect.x, rect.y, rect.width, rect.height)

    def contains(self, const int x, const int y, const int w, const int h) -> bool:
        return self.x<=x and self.y<=y and self.x+self.width>=x+w and self.y+self.height>=y+h

    def contains_rect(self, rectangle rect) -> bool:
        return self.contains(rect.x, rect.y, rect.width, rect.height)

    def subtract(self, const int x, const int y, const int w, const int h) -> list:
        """
        returns the rectangle(s) remaining when
        one subtracts the given rectangle from it
        """
        if w==0 or h==0 or self.width==0 or self.height==0:
            #no rectangle, no change:
            return [self]
        if self.x+self.width<=x or self.y+self.height<=y or x+w<=self.x or y+h<=self.y:
            #no intersection, no change:
            return [self]
        if x<=self.x and y<=self.y and x+w>=self.x+self.width and y+h>=self.y+self.height:
            #area contains this rectangle, so nothing remains:
            return []
        cdef object rects = []
        #note: we do "width first", no redundant area
        #which means we prefer wider rectangles for the areas that would overlap (the corners)
        if self.y<y:
            #top:
            rects.append(rectangle(self.x, self.y, self.width, y-self.y))
        #height for both sides:
        cdef int sy = MAX(self.y, y)
        cdef int sh = MIN(self.y+self.height, y+h)-sy
        cdef int nsx, nsw
        if sh>0:
            if self.x<x:
                #left:
                nsx = self.x
                nsw = x-nsx
                rects.append(rectangle(nsx, sy, nsw, sh))
            if self.x+self.width>x+w:
                #right:
                nsx = x+w
                nsw = self.x+self.width-(x+w)
                rects.append(rectangle(nsx, sy, nsw, sh))
        if self.y+self.height>y+h:
            #bottom:
            rects.append(rectangle(self.x, y+h, self.width, self.y+self.height-(y+h)))
        return rects

    def subtract_rect(self, rectangle rect) -> list:
        return self.subtract(rect.x, rect.y, rect.width, rect.height)

    def get_geometry(self) -> tuple[int, int, int, int]:
        return (self.x, self.y, self.width, self.height)

    def clone(self):
        return rectangle(self.x, self.y, self.width, self.height)


def contains(object regions, const int x, const int y, const int w, const int h) -> bool:
    cdef int x2 = x+w
    cdef int y2 = y+h
    for r in regions:
        if x>=r.x and y>=r.y and x2<=(r.x+r.width) and y2<=(r.y+r.height):
            return True
    return False


def contains_rect(object regions, rectangle region) -> bool:
    return contains(regions, region.x, region.y, region.width, region.height)


def add_rectangle(object regions, rectangle region) -> int:
    #returns the number of pixels actually added
    cdef int x = region.x
    cdef int y = region.y
    cdef int w = region.width
    cdef int h = region.height
    cdef long total
    cdef rectangle r, sub
    for r in tuple(regions):
        #unroll contains() call:
        #if r.contains_rect(region):
        if r.x<=x and r.y<=y and r.x+r.width>=x+w and r.y+r.height>=y+h:
            #rectangle is contained in another rectangle,
            #so no need to add anything
            return 0
        if r.intersects(x, y, w, h):
            total = 0
            #only add the parts that are not already in the rectangle
            #it intersects:
            for sub in region.subtract_rect(r):
                total += add_rectangle(regions, sub)
            return total
    #not found at all, add it all:
    regions.append(region)
    return w*h


def remove_rectangle(object regions, rectangle region) -> None:
    copy = regions[:]
    cdef int x = region.x               #
    cdef int y = region.y               #
    cdef int w = region.width           #
    cdef int h = region.height          #
    cdef rectangle r
    new_regions = []
    for r in copy:
        new_regions += r.subtract(x, y, w, h)
    regions[:] = new_regions


def merge_all(rectangles) -> rectangle:
    if not rectangles:
        raise ValueError("no rectangles to merge")
    cdef rectangle r = rectangles[0]
    cdef int rx = r.x
    cdef int ry = r.y
    cdef int rx2 = r.x + r.width
    cdef int ry2 = r.y + r.height
    cdef int x2, y2
    for r in rectangles:
        if not r:
            continue
        if r.x<rx:
            rx = r.x
        if r.y<ry:
            ry = r.y
        x2 = r.x + r.width
        y2 = r.y + r.height
        if x2>rx2:
            rx2 = x2
        if y2>ry2:
            ry2 = y2
    return rectangle(rx, ry, rx2-rx, ry2-ry)



# This file is part of Xpra.
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from io import BytesIO
from collections.abc import Sequence

from AppKit import (
    NSStringPboardType, NSTIFFPboardType, NSPasteboardTypePNG, NSPasteboardTypeURL,
    NSPasteboard,
)
from CoreFoundation import NSData, CFDataGetBytes, CFDataGetLength

from xpra.clipboard.timeout import ClipboardTimeoutHelper
from xpra.clipboard.core import _filter_targets, ClipboardProxyCore, TEXT_TARGETS, ClipboardCallback
from xpra.platform.ui_thread_watcher import get_UI_watcher
from xpra.util.str_fn import csv, Ellipsizer, bytestostr
from xpra.os_util import gi_import
from xpra.log import Logger

log = Logger("clipboard", "osx")

GLib = gi_import("GLib")

TARGET_TRANS = {
    NSStringPboardType: "STRING",
}

IMAGE_FORMATS = ["image/png", "image/jpeg", "image/tiff"]


def filter_targets(targets) -> Sequence[str]:
    return _filter_targets(TARGET_TRANS.get(x, x) for x in targets)


class OSXClipboardProxy(ClipboardProxyCore):

    def __init__(self, selection, pasteboard, send_clipboard_request_handler, send_clipboard_token_handler):
        self.pasteboard = pasteboard
        self.send_clipboard_request_handler = send_clipboard_request_handler
        self.send_clipboard_token_handler = send_clipboard_token_handler
        super().__init__(selection)
        self.update_change_count()
        # setup clipboard counter watcher:
        w = get_UI_watcher()
        w.add_alive_callback(self.timer_clipboard_check)

    def cleanup(self) -> None:
        super().cleanup()
        w = get_UI_watcher()
        if w:
            try:
                w.remove_alive_callback(self.timer_clipboard_check)
            except (KeyError, ValueError):
                pass

    def timer_clipboard_check(self) -> None:
        c = self.change_count
        self.update_change_count()
        log("timer_clipboard_check() was %s, now %s (have token: %s)", c, self.change_count, self._have_token)
        if c != self.change_count:
            self.local_clipboard_changed()

    def update_change_count(self) -> None:
        p = self.pasteboard
        if p:
            self.change_count = p.changeCount()

    def clear(self) -> None:
        self.pasteboard.clearContents()

    def do_emit_token(self) -> None:
        packet_data: list[str | int | list[str]] = []
        if self._want_targets:
            targets = self.get_targets()
            log("do_emit_token() targets=%s", targets)
            packet_data.append(targets)
            if self._greedy_client and "TEXT" in targets:
                text = self.get_clipboard_text()
                if text:
                    packet_data += ["STRING", "bytes", 8, text]
        self.send_clipboard_token_handler(self, tuple(packet_data))

    def get_clipboard_text(self) -> str:
        text = self.pasteboard.stringForType_(NSStringPboardType)
        log("get_clipboard_text() NSStringPboardType='%s' (%s)", text, type(text))
        return str(text)

    def get_targets(self) -> list[str]:
        types = self.pasteboard.types()
        targets = []
        if any(t in (NSStringPboardType, NSPasteboardTypeURL, "public.utf8-plain-text", "public.html", "TEXT") for t in
               types):
            targets += ["TEXT", "STRING", "text/plain", "text/plain;charset=utf-8", "UTF8_STRING"]
        if any(t in (NSTIFFPboardType, NSPasteboardTypePNG) for t in types):
            targets += IMAGE_FORMATS
        log("get_targets() targets(%s)=%s", types, targets)
        return targets

    def get_contents(self, target: str, got_contents: ClipboardCallback) -> None:
        log("get_contents%s", (target, got_contents))
        if target == "TARGETS":
            got_contents("ATOM", 32, self.get_targets())
            return
        if target in IMAGE_FORMATS:
            try:
                data = self.get_image_contents(target)
                if data:
                    got_contents(target, 8, data)
                    return
            except Exception:
                log.error("Error: failed to copy image from clipboard", exc_info=True)
        if target in ("TEXT", "STRING", "text/plain", "text/plain;charset=utf-8", "UTF8_STRING"):
            text = self.get_clipboard_text()
            got_contents(target, 8, text)
            return
        # we don't know how to handle this target,
        # return an empty response:
        got_contents(target, 8, b"")

    def get_image_contents(self, target):
        types = filter_targets(self.pasteboard.types())
        if target == "image/png" and NSPasteboardTypePNG in types:
            src_dtype = target
            img_data = self.pasteboard.dataForType_(NSPasteboardTypePNG)
        elif target == "image/tiff" and NSTIFFPboardType in types:
            src_dtype = target
            img_data = self.pasteboard.dataForType_(NSTIFFPboardType)
        elif NSPasteboardTypePNG in types:
            src_dtype = "image/png"
            img_data = self.pasteboard.dataForType_(NSPasteboardTypePNG)
        elif NSTIFFPboardType in types:
            src_dtype = "image/tiff"
            img_data = self.pasteboard.dataForType_(NSTIFFPboardType)
        else:
            log("image target '%s' not found in %s", target, types)
            return None
        if not img_data:
            return None
        length = CFDataGetLength(img_data)
        img_data = CFDataGetBytes(img_data, (0, length), None)
        img_data = self.filter_data(dtype=src_dtype, dformat=8, data=img_data, trusted=False, output_dtype=target)
        log("get_image_contents(%s)=%i %s", target, len(img_data or ()), type(img_data))
        return img_data

    def got_token(self, targets, target_data=None, claim=True, _synchronous_client=False) -> None:
        # the remote end now owns the clipboard
        self.cancel_emit_token()
        if not self._enabled:
            return
        self._got_token_events += 1
        log("got token, selection=%s, targets=%s, target data=%s, claim=%s, can-receive=%s",
            self._selection, targets, Ellipsizer(target_data), claim, self._can_receive)
        if self._can_receive:
            self.targets = _filter_targets(targets or ())
            self.target_data = target_data or {}
            if targets:
                self.got_contents("TARGETS", "ATOM", 32, targets)
            if target_data:
                for target, td_def in target_data.items():
                    dtype, dformat, data = td_def
                    dtype = bytestostr(dtype)
                    self.got_contents(target, dtype, dformat, data)
            # since we claim to be greedy
            # the peer should have sent us the target and target_data,
            # if not then request it:
            if not targets:
                self.send_clipboard_request_handler(self, self._selection, "TARGETS")
        if not claim:
            log("token packet without claim, not setting the token flag")
            return
        if claim:
            self._have_token = True

    def got_contents(self, target, dtype="", dformat: int = 8, data=b"") -> None:
        # if this is the special target 'TARGETS', cache the result:
        if target == "TARGETS" and dtype == "ATOM" and dformat == 32:
            self.targets = _filter_targets(data)
            log("got_contents: tell OS we have %s", csv(self.targets))
            image_types = tuple(t for t in IMAGE_FORMATS if t in self.targets)
            log("image_types=%s, dtype=%s (is text=%s)",
                image_types, dtype, dtype in TEXT_TARGETS)
            if image_types and dtype not in TEXT_TARGETS:
                # request image:
                self.send_clipboard_request_handler(self, self._selection, image_types[0])
            return
        if dformat == 8 and dtype in TEXT_TARGETS:
            log("we got a byte string: %s", Ellipsizer(data))
            self.set_clipboard_text(bytestostr(data))
        if dformat == 8 and dtype in IMAGE_FORMATS:
            log("we got a %s image", dtype)
            self.set_image_data(dtype, data)

    def set_image_data(self, dtype, data):
        img_type = dtype.split("/")[1]  # ie: "png"
        from xpra.codecs.pillow.decoder import open_only
        img = open_only(data, (img_type,))
        for img_type, macos_types in {
            "png": [NSPasteboardTypePNG, "image/png"],
            "tiff": [NSTIFFPboardType, "image/tiff"],
            "jpeg": ["public.jpeg", "image/jpeg"],
        }.items():
            try:
                save_img = img
                if img_type == "jpeg" and img.mode == "RGBA":
                    save_img = img.convert("RGB")
                buf = BytesIO()
                save_img.save(buf, img_type)
                data = buf.getvalue()
                buf.close()
                self.pasteboard.clearContents()
                nsdata = NSData.dataWithData_(data)
                for t in macos_types:
                    r = self.pasteboard.setData_forType_(nsdata, t)
                    log("set '%s' data type: %s", t, r)
            except Exception as e:
                log("set_image_data(%s, ..)", dtype, exc_info=True)
                log.error("Error: failed to copy %s image to clipboard", img_type)
                log.estr(e)

    def set_clipboard_text(self, text) -> None:
        self.pasteboard.clearContents()
        r = self.pasteboard.setString_forType_(text, NSStringPboardType)
        log("set_clipboard_text(%s) success=%s", text, r)
        self.update_change_count()

    def local_clipboard_changed(self) -> None:
        log("local_clipboard_changed()")
        self.do_owner_changed()


class OSXClipboardProtocolHelper(ClipboardTimeoutHelper):

    def __init__(self, *args, **kwargs):
        self.pasteboard = NSPasteboard.generalPasteboard()
        if self.pasteboard is None:
            raise RuntimeError("cannot load Pasteboard, maybe not running from a GUI session?")
        kwargs["clipboard.local"] = "CLIPBOARD"
        kwargs["clipboards.local"] = ["CLIPBOARD"]
        super().__init__(*args, **kwargs)

    def __repr__(self):
        return "OSXClipboardProtocolHelper"

    def cleanup(self) -> None:
        super().cleanup()
        self.pasteboard = None

    def make_proxy(self, selection) -> OSXClipboardProxy:
        proxy = OSXClipboardProxy(selection, self.pasteboard,
                                  self._send_clipboard_request_handler, self._send_clipboard_token_handler)
        proxy.set_direction(self.can_send, self.can_receive)
        return proxy

    ############################################################################
    # just pass ATOM targets through
    # (we use them internally as strings)
    ############################################################################
    def _munge_wire_selection_to_raw(self, encoding, dtype, dformat, data) -> bytes | str:
        if encoding == "atoms":
            data = _filter_targets(data)
        return super()._munge_wire_selection_to_raw(encoding, dtype, dformat, data)


def main():
    import time
    from xpra.platform import program_context
    with program_context("OSX Clipboard Change Test"):
        log.enable_debug()

        # init UI watcher with gobject (required by pasteboard monitoring code)
        get_UI_watcher()

        log.info("testing pasteboard")

        gtk = gi_import("Gtk")
        pasteboard = NSPasteboard.generalPasteboard()

        def nosend(*args):
            log("nosend%s", args)

        proxy = OSXClipboardProxy("CLIPBOARD", pasteboard, nosend, nosend)
        log.info("current change count=%s", proxy.change_count)
        clipboard = gtk.Clipboard(selection="CLIPBOARD")
        log.info("changing clipboard %s contents", clipboard)
        clipboard.set_text("HELLO WORLD %s" % time.time())
        proxy.update_change_count()
        log.info("new change count=%s", proxy.change_count)
        log.info("any update to your clipboard should get logged (^C to exit)")
        cc = proxy.change_count
        while True:
            v = proxy.change_count
            if v != cc:
                log.info("success! the clipboard change has been detected, new change count=%s", v)
            else:
                log.info(".")
            time.sleep(1)


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from typing import Any
from collections.abc import Sequence

from xpra.util.objects import typedict
from xpra.server.source.stub_source_mixin import StubSourceMixin
from xpra.net.mmap import init_server_mmap, BaseMmapArea

from xpra.log import Logger

log = Logger("mmap")


def clean_mmap_area(area: BaseMmapArea) -> None:
    if not area:
        return
    try:
        area.close()
    except BufferError:
        # this can fail when handling a SIGINT signal,
        # with the message: "cannot close exported pointers exist",
        # so wait a little bit and try again:
        from xpra.os_util import gi_import
        GLib = gi_import("GLib")

        def retry_close():
            try:
                area.close()
            except RuntimeError as e:
                log.warn(f"Warning: failed to close mmap area {area!r}: {e}")
        GLib.timeout_add(100, retry_close)


class MMAP_Connection(StubSourceMixin):

    PREFIX = "mmap"

    @classmethod
    def is_needed(cls, caps: typedict) -> bool:
        mmap_caps = typedict(caps.get("mmap") or {})
        for prefix in ("", "read", "write"):
            if prefix:
                area_caps = typedict(mmap_caps.dictget(prefix) or {})
            else:
                # older versions supply a single unprefixed area:
                area_caps = mmap_caps
            log(f"is_needed(..) caps({prefix!r})={area_caps!r}")
            if area_caps.intget("size", 0) > 0:
                return True
        return False

    def __init__(self):
        self.mmap_supported = False
        self.mmap_read_area = None
        self.mmap_write_area = None
        self.mmap_min_size = 0
        self.mmap_filenames: Sequence[str] = ()

    def init_from(self, _protocol, server) -> None:
        self.mmap_supported = server.mmap_supported
        if server.mmap_filename:
            self.mmap_filenames = server.mmap_filename.split(os.path.pathsep)
        self.mmap_min_size = server.mmap_min_size

    def init_state(self) -> None:
        self.mmap_read_area = None
        self.mmap_write_area = None

    def cleanup(self) -> None:
        clean_mmap_area(self.mmap_read_area)
        self.mmap_read_area = None
        clean_mmap_area(self.mmap_write_area)
        self.mmap_write_area = None

    def mmap_path(self, filename: str, index: int) -> str:
        if len(self.mmap_filenames) == 1 and os.path.isdir(self.mmap_filenames[0]):
            # server directory specified: use the client's filename, but at the server path
            mmap_dir = self.mmap_filenames[0]
            log(f"using global server specified mmap directory: {mmap_dir!r}")
            return os.path.join(mmap_dir, os.path.basename(filename))
        if self.mmap_filenames and len(self.mmap_filenames) > index:
            # server command line option overrides the path:
            filename = self.mmap_filenames[index]
            log(f"using global server specified mmap file path: {filename!r}")
        return filename

    def parse_area_caps(self, name: str, raw_caps: dict, index: int) -> BaseMmapArea | None:
        log("parse_area_caps(%r, %r, %i)", name, raw_caps, index)
        if not raw_caps:
            return None
        caps = typedict(raw_caps)
        filename = self.mmap_path(caps.strget("file"), index)
        if not filename or not os.path.exists(filename):
            log(f"mmap_file {filename!r} cannot be found!")
            return None
        size = caps.intget("size", 0)
        log("client supplied mmap_file=%r, size=%i", filename, size)
        if not size:
            return None
        area = BaseMmapArea(name, filename, size)
        area.parse_caps(caps)
        if not area.enabled:
            return None
        mmap, size = init_server_mmap(filename, size)
        log("found client mmap area: %s, %i bytes - min mmap size=%i in %r",
            mmap, size, self.mmap_min_size, filename)
        if size <= 0 or not mmap:
            return None
        if size < self.mmap_min_size:
            mmap.close()
            log.warn("Warning: client %s supplied mmap area is too small, discarding it", name)
            log.warn(" at least %iMB are needed and this area is only %iMB",
                     self.mmap_min_size // 1024 // 1024, size // 1024 // 1024)
            return None
        area.mmap = mmap
        area.size = size
        if not area.verify_token():
            mmap.close()
            return None
        from xpra.util.stats import std_unit
        log.info(" mmap is enabled using %sB area in %s", std_unit(size, unit=1024), filename)
        return area

    def parse_client_caps(self, caps: typedict) -> None:
        if not self.mmap_supported:
            log("mmap.parse_client_caps() mmap is disabled")
            return
        mmap_caps = caps.get("mmap") or {}
        if not mmap_caps or not isinstance(mmap_caps, dict):
            log(f"mmap.parse_client_caps() client did not supply valid mmap caps: {mmap_caps!r}")
            self.mmap_supported = False
            return
        tdcaps = typedict(mmap_caps)
        self.mmap_read_area = self.parse_area_caps("read", tdcaps.dictget("write") or mmap_caps, 1)
        # also try legacy unprefixed lookup for 'read' area:
        self.mmap_write_area = self.parse_area_caps("write", tdcaps.dictget("read"), 0)
        log("parse_client_caps() mmap-read=%s, mmap-write=%s", self.mmap_read_area, self.mmap_write_area)

    def get_caps(self) -> dict[str, Any]:
        mmap_caps: dict[str, Any] = {}
        for prefixes, area in (
            (("read", ""), self.mmap_read_area),
            (("write", ), self.mmap_write_area),
        ):
            if not area:
                continue
            # write a new token that the client can verify:
            area.gen_token()
            area.write_token()
            caps = area.get_caps()
            for prefix in prefixes:
                if prefix:
                    mmap_caps[prefix] = caps
                else:
                    mmap_caps.update(caps)
        log(f"mmap caps={mmap_caps}")
        return {MMAP_Connection.PREFIX: mmap_caps}

    def get_info(self) -> dict[str, Any]:
        info = {}
        for name, area in {
            "read": self.mmap_read_area,
            "write": self.mmap_write_area,
        }.items():
            if area:
                info[name] = area.get_info()
        return {MMAP_Connection.PREFIX: info}



# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from libc.stdint cimport uint8_t, uint32_t
from xpra.buffers.membuf cimport MemBuf, getbuf
from typing import Tuple, Dict

from xpra.common import MAX_DECOMPRESSED_SIZE
from xpra.log import Logger
log = Logger("brotli")


cdef extern from "Python.h":
    int PyObject_GetBuffer(object obj, Py_buffer *view, int flags)
    void PyBuffer_Release(Py_buffer *view)
    int PyBUF_ANY_CONTIGUOUS

cdef extern from "brotli/decode.h":
    ctypedef enum BrotliDecoderResult:
        # Decoding error, e.g. corrupted input or memory allocation problem
        BROTLI_DECODER_RESULT_ERROR
        # Decoding successfully completed
        BROTLI_DECODER_RESULT_SUCCESS
        # Partially done; should be called again with more input
        BROTLI_DECODER_RESULT_NEEDS_MORE_INPUT
        # Partially done; should be called again with more output
        BROTLI_DECODER_RESULT_NEEDS_MORE_OUTPUT

    ctypedef void BrotliDecoderState
    ctypedef void* brotli_alloc_func
    ctypedef void* brotli_free_func

    ctypedef int BrotliDecoderErrorCode

    BrotliDecoderErrorCode BrotliDecoderGetErrorCode(const BrotliDecoderState* state)
    const char* BrotliDecoderErrorString(BrotliDecoderErrorCode c)
    uint32_t BrotliDecoderVersion()

    BrotliDecoderState* BrotliDecoderCreateInstance(brotli_alloc_func alloc_func,
                                                    brotli_free_func free_func,
                                                    void* opaque)

    void BrotliDecoderDestroyInstance(BrotliDecoderState* state)

    # Decompresses the input stream to the output stream
    BrotliDecoderResult BrotliDecoderDecompressStream(BrotliDecoderState * state,
                                                      size_t * available_in, const uint8_t ** next_in,
                                                      size_t * available_out, uint8_t ** next_out,
                                                      size_t * total_out) nogil

    int BrotliDecoderHasMoreOutput(const BrotliDecoderState* state)
    const uint8_t* BrotliDecoderTakeOutput(BrotliDecoderState* state, size_t* size)
    int BrotliDecoderIsUsed(const BrotliDecoderState* state)
    int BrotliDecoderIsFinished(const BrotliDecoderState* state)


RESULT_STR : Dict[BrotliDecoderResult, str] = {
    BROTLI_DECODER_RESULT_ERROR : "error",
    BROTLI_DECODER_RESULT_SUCCESS   : "success",
    BROTLI_DECODER_RESULT_NEEDS_MORE_INPUT : "needs-more-input",
    BROTLI_DECODER_RESULT_NEEDS_MORE_OUTPUT : "needs-more-output",
}


def get_version() -> Tuple[int, int, int]:
    cdef uint32_t bv = BrotliDecoderVersion()
    cdef unsigned int major = bv >> 24
    cdef unsigned int minor = (bv >> 12) & 0xFFF
    cdef unsigned int patch = bv & 0xFFF
    return (major, minor, patch)


def decompress(data, maxsize=MAX_DECOMPRESSED_SIZE) -> bytes:
    cdef const uint8_t *in_ptr = NULL
    cdef size_t available_in = 0
    cdef MemBuf out_buf = getbuf(512*1024, True)
    cdef size_t available_out = 0
    cdef uint8_t *out_ptr = NULL
    cdef size_t total_out = 0
    cdef size_t decoded = 0

    cdef BrotliDecoderState* state = NULL
    cdef Py_buffer in_buf
    cdef BrotliDecoderResult r
    chunks = []

    if PyObject_GetBuffer(data, &in_buf, PyBUF_ANY_CONTIGUOUS):
        raise ValueError("failed to read data from %s" % type(data))
    in_ptr = <const uint8_t*> in_buf.buf
    available_in = in_buf.len
    log("brotli.decompress(%i bytes, %i)", available_in, maxsize)
    try:
        state = BrotliDecoderCreateInstance(NULL, NULL, NULL)
        assert state!=NULL, "failed to allocate a brotli decoder instance"
        while True:
            available_out = len(out_buf)
            out_ptr = <uint8_t*> out_buf.get_mem()
            with nogil:
                r = BrotliDecoderDecompressStream(state,
                                                  &available_in, &in_ptr,
                                                  &available_out, &out_ptr,
                                                  &total_out)
            log("BrotliDecoderDecompressStream(..)=%s", RESULT_STR.get(r, r))
            if total_out>maxsize:
                raise ValueError("brotli decompression would exceed maximum size allowed %i" % maxsize)
            if r==BROTLI_DECODER_RESULT_ERROR:
                raise ValueError("brotli decoder error")
            if r==BROTLI_DECODER_RESULT_NEEDS_MORE_INPUT:
                if available_in<=0:
                    raise ValueError("brotli decompressor expected more input data")
                continue
            assert r==BROTLI_DECODER_RESULT_NEEDS_MORE_OUTPUT or r==BROTLI_DECODER_RESULT_SUCCESS, "unknown return value %i" % r
            decoded = len(out_buf) - available_out
            assert decoded>0, "decoder returned no data"
            out_ptr = <uint8_t*> out_buf.get_mem()
            b = out_ptr[:decoded]
            chunks.append(b)
            if r==BROTLI_DECODER_RESULT_SUCCESS:
                break
    finally:
        PyBuffer_Release(&in_buf)
        if state:
            BrotliDecoderDestroyInstance(state)
    del out_buf
    if len(chunks) == 1:
        return chunks[0]
    return b"".join(chunks)



# This file is part of Xpra.
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys


def main():
    from xpra.platform import program_context
    with program_context("Print-Atoms", "Print Atoms"):
        from xpra.x11.gtk import gdk_display_source  # @UnresolvedImport, @Reimport

        gdk_display_source.init_gdk_display_source()  # @UndefinedVariable

        from xpra.x11.bindings.core import X11CoreBindings  # @UnresolvedImport

        X11Core = X11CoreBindings()

        for s in sys.argv[1:]:
            if s.lower().startswith("0x"):
                v = int(s, 16)
            else:
                v = int(s)
            print("%s : %s" % (s, X11Core.get_atom_name(v)))


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import struct
from typing import Dict
from collections import deque

from xpra.log import Logger
log = Logger("x11", "bindings", "xinput")

from xpra.gtk.error import xlog
from xpra.x11.common import X11Event
from xpra.util.str_fn import hexstr

from libc.string cimport memset  # pylint: disable=syntax-error
from xpra.x11.bindings.xlib cimport (
    Display, Bool, Time, Window, Atom, Status,
    XGenericEventCookie,
    XQueryExtension,
    XGetEventData, XFreeEventData, XDefaultRootWindow, XQueryPointer,
    XFlush,
    XInternAtom, XFree,
    BadRequest, Success, XIAnyPropertyType,
)
from libc.stdint cimport uintptr_t   # pylint: disable=syntax-error


######
# Xlib primitives and constants
######

ctypedef unsigned long CARD32
DEF XNone = 0

cdef extern from "X11/extensions/XInput2.h":
    int XI_LASTEVENT
    int XI_DeviceChanged
    int XI_KeyPress
    int XI_KeyRelease
    int XI_ButtonPress
    int XI_ButtonRelease
    int XI_Motion
    int XI_Enter
    int XI_Leave
    int XI_FocusIn
    int XI_FocusOut
    int XI_HierarchyChanged
    int XI_PropertyEvent
    int XI_RawKeyPress
    int XI_RawKeyRelease
    int XI_RawButtonPress
    int XI_RawButtonRelease
    int XI_RawMotion
    int XI_TouchBegin
    int XI_TouchUpdate
    int XI_TouchEnd
    int XI_TouchOwnership
    int XI_RawTouchBegin
    int XI_RawTouchUpdate
    int XI_RawTouchEnd

    int XIMasterPointer
    int XIMasterKeyboard
    int XISlavePointer
    int XISlaveKeyboard
    int XIFloatingSlave

    int XIButtonClass
    int XIKeyClass
    int XIValuatorClass
    int XIScrollClass
    int XITouchClass

    int XIAllDevices
    int XIAllMasterDevices

    ctypedef struct XIValuatorState:
        int           mask_len
        unsigned char *mask
        double        *values

    ctypedef struct XIEvent:
        int           type
        unsigned long serial
        Bool          send_event
        Display       *display
        int           extension
        int           evtype
        Time          time

    ctypedef struct XIRawEvent:
        int           type      #GenericEvent
        unsigned long serial
        Bool          send_event
        Display       *display
        int           extension #XI extension offset
        int           evtype    #XI_RawKeyPress, XI_RawKeyRelease, etc
        Time          time
        int           deviceid
        int           sourceid
        int           detail
        int           flags
        XIValuatorState valuators
        double        *raw_values

    ctypedef struct XIButtonState:
        int           mask_len
        unsigned char *mask

    ctypedef struct XIModifierState:
        int    base
        int    latched
        int    locked
        int    effective

    ctypedef XIModifierState XIGroupState

    ctypedef struct XIDeviceEvent:
        int           type
        unsigned long serial
        Bool          send_event
        Display       *display
        int           extension
        int           evtype
        Time          time
        int           deviceid
        int           sourceid
        int           detail
        Window        root
        Window        event
        Window        child
        double        root_x
        double        root_y
        double        event_x
        double        event_y
        int           flags
        XIButtonState       buttons
        XIValuatorState     valuators
        XIModifierState     mods
        XIGroupState        group

    ctypedef struct XIHierarchyInfo:
        int           deviceid
        int           attachment
        int           use
        Bool          enabled
        int           flags

    ctypedef struct XIHierarchyEvent:
        int           type
        unsigned long serial
        Bool          send_event
        Display       *display
        int           extension
        int           evtype            #XI_HierarchyChanged
        Time          time
        int           flags
        int           num_info
        XIHierarchyInfo *info

    ctypedef struct XIEventMask:
        int                 deviceid
        int                 mask_len
        unsigned char*      mask

    ctypedef struct XIAnyClassInfo:
        int         type
        int         sourceid

    ctypedef struct XIDeviceInfo:
        int                 deviceid
        char                *name
        int                 use
        int                 attachment
        Bool                enabled
        int                 num_classes
        XIAnyClassInfo      **classes

    ctypedef struct XIButtonClassInfo:
        int         type
        int         sourceid
        int         num_buttons
        Atom        *labels
        XIButtonState state

    ctypedef struct XIKeyClassInfo:
        int         type
        int         sourceid
        int         num_keycodes
        int         *keycodes

    ctypedef struct XIValuatorClassInfo:
        int         type
        int         sourceid
        int         number
        Atom        label
        double      min
        double      max
        double      value
        int         resolution
        int         mode

    ctypedef struct XIScrollClassInfo:
        int         type
        int         sourceid
        int         number
        int         scroll_type
        double      increment
        int         flags

    ctypedef struct XITouchClassInfo:
        int         type
        int         sourceid
        int         mode
        int         num_touches

    Status XIQueryVersion(Display *display, int *major_version_inout, int *minor_version_inout)
    Status XISelectEvents(Display *display, Window win, XIEventMask *masks, int num_masks)
    XIDeviceInfo* XIQueryDevice(Display *display, int deviceid, int *ndevices_return)
    void XIFreeDeviceInfo(XIDeviceInfo *info)
    Atom *XIListProperties(Display *display, int deviceid, int *num_props_return)
    Status XIGetProperty(Display *display, int deviceid, Atom property, long offset, long length,
                         Bool delete_property, Atom type, Atom *type_return,
                         int *format_return, unsigned long *num_items_return,
                         unsigned long *bytes_after_return, unsigned char **data)


DEF MAX_XI_EVENTS = 64
DEF XI_EVENT_MASK_SIZE = (MAX_XI_EVENTS+7)//8

XI_EVENT_NAMES = {
    XI_DeviceChanged    : "XI_DeviceChanged",
    XI_KeyPress         : "XI_KeyPress",
    XI_KeyRelease       : "XI_KeyRelease",
    XI_ButtonPress      : "XI_ButtonPress",
    XI_ButtonRelease    : "XI_ButtonRelease",
    XI_Motion           : "XI_Motion",
    XI_Enter            : "XI_Enter",
    XI_Leave            : "XI_Leave",
    XI_FocusIn          : "XI_FocusIn",
    XI_FocusOut         : "XI_FocusOut",
    XI_HierarchyChanged : "XI_HierarchyChanged",
    XI_PropertyEvent    : "XI_PropertyEvent",
    XI_RawKeyPress      : "XI_RawKeyPress",
    XI_RawKeyRelease    : "XI_RawKeyRelease",
    XI_RawButtonPress   : "XI_RawButtonPress",
    XI_RawButtonRelease : "XI_RawButtonRelease",
    XI_RawMotion        : "XI_RawMotion",
    XI_TouchBegin       : "XI_TouchBegin",
    XI_TouchUpdate      : "XI_TouchUpdate",
    XI_TouchEnd         : "XI_TouchEnd",
    XI_TouchOwnership   : "XI_TouchOwnership",
    XI_RawTouchBegin    : "XI_RawTouchBegin",
    XI_RawTouchUpdate   : "XI_RawTouchUpdate",
    XI_RawTouchEnd      : "XI_RawTouchEnd",
}

XI_USE = {
    XIMasterPointer     : "master pointer",
    XIMasterKeyboard    : "master keyboard",
    XISlavePointer      : "slave pointer",
    XISlaveKeyboard     : "slave keyboard",
    XIFloatingSlave     : "floating slave",
}

CLASS_INFO = {
    XIButtonClass       : "button",
    XIKeyClass          : "key",
    XIValuatorClass     : "valuator",
    XIScrollClass       : "scroll",
    XITouchClass        : "touch",
}


from xpra.x11.bindings.core cimport X11CoreBindingsInstance


cdef class X11XI2BindingsInstance(X11CoreBindingsInstance):

    cdef int opcode
    cdef object events
    cdef object event_handlers
    cdef object device_cache

    def __init__(self):
        self.opcode = -1
        self.event_handlers = {}
        self.device_cache = {}
        self.reset_events()

    def __repr__(self):
        return "X11XI2Bindings(%s)" % self.display_name

    def connect(self, window, event, handler) -> None:
        self.event_handlers.setdefault(window, {})[event] = handler

    def disconnect(self, window) -> None:
        try:
            del self.event_handlers[window]
        except:
            pass

    def reset_events(self) -> None:
        self.events = deque(maxlen=100)

    def find_event(self, event_name: str, serial: int) -> int:
        for x in reversed(self.events):
            #log.info("find_event(%s, %#x) checking %s", event_name, serial, x)
            if x.name==event_name and x.serial==serial:
                #log.info("matched")
                return x
            if x.serial<serial:
                #log.info("serial too old")
                return 0
        return 0

    def find_events(self, event_name: str, windows) -> List[int]:
        cdef Window found = 0
        cdef Window window
        matches = []
        for x in reversed(self.events):
            window = x.xid
            if x.name==event_name and ((found>0 and found==window) or (found==0 and window in windows)):
                matches.append(x)
                found = window
            elif found:
                break
        return matches

    def get_xi_version(self, int major=2, int minor=2) -> Tuple[int, int]:
        self.context_check("get_xi_version")
        cdef int rmajor = major, rminor = minor
        cdef int rc = XIQueryVersion(self.display, &rmajor, &rminor)
        if rc == BadRequest:
            log.warn("Warning: no XI2 %i.%i support,", major, minor)
            log.warn(" server supports version %i.%i only", rmajor, rminor)
            return 0, 0
        log("get_xi_version%s=%s", (major, minor), (rmajor, rminor))
        return rmajor, rminor

    cdef int get_xi_opcode(self, int major=2, int minor=2):
        if self.opcode!=-1:
            return self.opcode
        cdef int opcode, event, error
        if not XQueryExtension(self.display, "XInputExtension", &opcode, &event, &error):
            log.warn("Warning: XI2 events are not supported")
            self.opcode = 0
            return 0
        cdef int rmajor = major, rminor = minor
        cdef int rc = XIQueryVersion(self.display, &rmajor, &rminor)
        if rc == BadRequest:
            log.warn("Warning: no XI2 %i.%i support,", major, minor)
            log.warn(" server supports version %i.%i only", rmajor, rminor)
            self.opcode = 0
            return 0
        elif rc:
            log.warn("Warning: Xlib bug querying XI2, code %i", rc)
            self.opcode = 0
            return 0
        self.opcode = opcode
        log("get_xi_opcode%s=%i", (major, minor), opcode)
        return opcode

    cdef register_parser(self):
        log("register_parser()")
        if self.opcode>0:
            from xpra.x11.bindings.events import add_x_event_parser
            add_x_event_parser(self.opcode, self.parse_xi_event)

    cdef register_gdk_events(self):
        log("register_gdk_events()")
        if self.opcode<=0:
            return
        global XI_EVENT_NAMES
        from xpra.x11.bindings.events import add_x_event_signal, add_x_event_type_name
        for e, xi_event_name in {
            XI_DeviceChanged    : "device-changed",
            XI_KeyPress         : "key-press",
            XI_KeyRelease       : "key-release",
            XI_ButtonPress      : "button-press",
            XI_ButtonRelease    : "button-release",
            XI_Motion           : "motion",
            XI_Enter            : "enter",
            XI_Leave            : "leave",
            XI_FocusIn          : "focus-in",
            XI_FocusOut         : "focus-out",
            XI_HierarchyChanged : "focus-changed",
            XI_PropertyEvent    : "property-event",
            XI_RawKeyPress      : "raw-key-press",
            XI_RawKeyRelease    : "raw-key-release",
            XI_RawButtonPress   : "raw-button-press",
            XI_RawButtonRelease : "raw-button-release",
            XI_RawMotion        : "raw-motion",
            XI_TouchBegin       : "touch-begin",
            XI_TouchUpdate      : "touch-update",
            XI_TouchEnd         : "touch-end",
            XI_TouchOwnership   : "touch-ownership",
            XI_RawTouchBegin    : "raw-touch-begin",
            XI_RawTouchUpdate   : "raw-touch-update",
            XI_RawTouchEnd      : "raw-touch-end",
        }.items():
            event = self.opcode+e
            add_x_event_signal(event, ("xi-%s" % xi_event_name, None))
            name = XI_EVENT_NAMES[e]
            add_x_event_type_name(event, name)

    def select_xi2_events(self) -> None:
        self.context_check("select_xi2_events")
        cdef Window win = XDefaultRootWindow(self.display)
        log("select_xi2_events() root window=%#x", win)
        assert XI_LASTEVENT<MAX_XI_EVENTS, "bug: source needs to be updated, XI_LASTEVENT=%i" % XI_LASTEVENT
        cdef XIEventMask evmasks[1]
        cdef unsigned char mask1[XI_EVENT_MASK_SIZE]
        memset(mask1, 0, XI_EVENT_MASK_SIZE)
        #define XISetMask(ptr, event)   (((unsigned char*)(ptr))[(event)>>3] |=  (1 << ((event) & 7)))
        #XISetMask(mask1, XI_RawMotion)
        for e in (
            XI_KeyPress, XI_KeyRelease,
            XI_Motion,
            XI_HierarchyChanged,
            XI_ButtonPress, XI_ButtonRelease,
            XI_RawButtonPress, XI_RawButtonRelease,
            XI_TouchBegin, XI_TouchUpdate, XI_TouchEnd,
            XI_RawTouchBegin, XI_RawTouchUpdate, XI_RawTouchEnd,
            XI_RawMotion,
            ):
            mask1[e>>3] |= (1<< (e & 0x7))
        evmasks[0].deviceid = XIAllDevices  #XIAllMasterDevices    #XIAllDevices
        evmasks[0].mask_len = XI_EVENT_MASK_SIZE
        evmasks[0].mask = mask1
        XISelectEvents(self.display, win, evmasks, 1)
        XFlush(self.display)

    def parse_xi_event(self, cookie):
        with xlog:
            return self.do_parse_xi_event(cookie)

    cdef do_parse_xi_event(self, uintptr_t _cookie):
        self.context_check("parse_xi_event")
        cdef XGenericEventCookie *cookie = <XGenericEventCookie*> _cookie
        cdef XIHierarchyEvent *hierarchy_e
        cdef XIRawEvent *raw
        cdef int i = 0
        if not XGetEventData(self.display, cookie):
            log("parse_xi_event(%#x) no event data", _cookie)
            return None
        cdef XIEvent *xie = <XIEvent*> cookie.data
        #cdef XIDeviceEvent *device_e = <XIDeviceEvent*> cookie.data
        cdef int xi_type = cookie.evtype
        etype = self.opcode+xi_type
        global XI_EVENT_NAMES
        event_name = XI_EVENT_NAMES.get(xi_type)
        if not event_name:
            log("parse_xi_event(%#x) unknown XI2 event code: %i", _cookie, xi_type)
            return None

        #don't parse the same thing again:
        if len(self.events)>0:
            last_event = self.events[-1]
            if last_event.serial==xie.serial and last_event.type==etype:
                log("parse_xi_event(%#x) repeated %s event skipped (%s)",
                    _cookie, last_event.name, event_name)
                return None

        pyev = X11Event(event_name)
        pyev.type = etype
        pyev.send_event = bool(xie.send_event)
        pyev.serial = xie.serial
        pyev.time = int(xie.time)
        pyev.window = int(XDefaultRootWindow(self.display))

        device_info = None
        if xi_type in (XI_KeyPress, XI_KeyRelease,
                       XI_ButtonPress, XI_ButtonRelease,
                       XI_Motion,
                       XI_TouchBegin, XI_TouchUpdate, XI_TouchEnd):
            device = <XIDeviceEvent*> cookie.data
            #pyev.source = device.sourceid    #always 0
            device_info = self.device_cache.get(device.deviceid)
            pyev.device = device.deviceid
            pyev.detail = device.detail
            pyev.flags = device.flags
            pyev.window = int(device.child or device.event or device.root)
            pyev.x_root = device.root_x
            pyev.y_root = device.root_y
            pyev.x = device.event_x
            pyev.y = device.event_y
            #mask = []
            valuators = {}
            valuator = 0
            for i in range(device.valuators.mask_len*8):
                if device.valuators.mask[i>>3] & (1 << (i & 0x7)):
                    valuators[i] = device.valuators.values[valuator]
                    valuator += 1
            pyev.valuators = valuators
            buttons = []
            for i in range(device.buttons.mask_len):
                if device.buttons.mask[i>>3] & (1<< (i & 0x7)):
                    buttons.append(i)
            pyev.buttons = buttons
            state = []
            pyev.state = state
            pyev.modifiers = {
                "base"      : device.mods.base,
                "latched"   : device.mods.latched,
                "locked"    : device.mods.locked,
                "effective" : device.mods.effective,
            }
            #make it compatible with gdk events:
            pyev.state = device.mods.effective
        elif xi_type in (XI_RawKeyPress, XI_RawKeyRelease,
                         XI_RawButtonPress, XI_RawButtonRelease,
                         XI_RawMotion,
                         XI_RawTouchBegin, XI_RawTouchUpdate, XI_RawTouchEnd):
            raw = <XIRawEvent*> cookie.data
            valuators = {}
            raw_valuators = {}
            valuator = 0
            for i in range(raw.valuators.mask_len*8):
                if raw.valuators.mask[i>>3] & (1 << (i & 0x7)):
                    valuators[i] = raw.valuators.values[valuator]
                    raw_valuators[i] = raw.raw_values[valuator]
                    valuator += 1
            pyev.valuators = valuators
            pyev.raw_valuators = raw_valuators
        elif xi_type == XI_HierarchyChanged:
            hierarchy_e = <XIHierarchyEvent*> cookie.data
            pyev.window = 0
            pyev.flags = hierarchy_e.flags
            #for i in range(hierarchy_e.num_info):
            #XIHierarchyInfo *info
        XFreeEventData(self.display, cookie)
        pyev.xid = pyev.window
        self.events.append(pyev)

        handler = self.event_handlers.get(pyev.window, {}).get(event_name)
        if handler:
            log("parse_xi_event: %s, device_info=%s, handler=%s", pyev, device_info, handler)
            handler(pyev, device_info)
        log("parse_xi_event: no handler for %s", event_name)
        return None

    def get_devices(self, show_all=True, show_disabled=False) -> Dict[int, Dict[str, Any]]:
        log("get_devices(%s, %s)", show_all, show_disabled)
        self.context_check("get_devices")
        global XI_USE
        cdef int ndevices, i, j

        cdef XIDeviceInfo *device
        cdef XIAnyClassInfo *clazz
        if show_all:
            device_types = XIAllDevices
        else:
            device_types = XIAllMasterDevices
        cdef XIDeviceInfo *devices = XIQueryDevice(self.display, device_types, &ndevices)
        dinfo = {}
        for i in range(ndevices):
            device = &devices[i]
            if not device.enabled and not show_disabled:
                continue
            info = {
                "name"          : device.name,
                "use"           : XI_USE.get(device.use, "unknown use: %i" % device.use),
                "attachment"    : device.attachment,
                "enabled"       : device.enabled,
                }
            classes = {}
            for j in range(device.num_classes):
                clazz = device.classes[j]
                classes[j] = self.get_class_info(clazz)
            info["classes"] = classes
            properties = self.get_device_properties(device.deviceid)
            if properties:
                info["properties"] = properties
            log("[%i] %s: %s", device.deviceid, device.name, info)
            dinfo[device.deviceid] = info
        XIFreeDeviceInfo(devices)
        self.device_cache = dinfo
        return dinfo

    def get_device_properties(self, deviceid) -> Dict[str, Any]:
        self.context_check("get_device_properties")
        cdef int nprops, i
        cdef Atom *atoms = XIListProperties(self.display, deviceid, &nprops)
        if atoms==NULL or nprops==0:
            return None
        props = {}
        for i in range(nprops):
            value = self.get_device_property(deviceid, atoms[i])
            if value is not None:
                prop_name = self.get_atom_name(atoms[i])
                props[prop_name] = value
        return props

    cdef get_device_property(self, int deviceid, Atom property, req_type=0):
        #code mostly duplicated from window bindings XGetWindowProperty:
        cdef int buffer_size = 64 * 1024
        cdef Atom xactual_type = <Atom> 0
        cdef int actual_format = 0
        cdef unsigned long nitems = 0, bytes_after = 0
        cdef unsigned char *prop = NULL
        cdef Atom xreq_type = XIAnyPropertyType
        if req_type:
            xreq_type = self.get_xatom(req_type)

        cdef Status status = XIGetProperty(self.display,
                               deviceid, property,
                               0,
                               buffer_size//4,
                               False,
                               xreq_type, &xactual_type,
                               &actual_format, &nitems, &bytes_after, &prop)
        if status != Success:
            raise RuntimeError("failed to retrieve XI property")
        if xactual_type == XNone:
            return None
        if xreq_type and xreq_type != xactual_type:
            raise RuntimeError("expected %s but got %s" % (req_type, self.get_atom_name(xactual_type)))
        # This should only occur for bad property types:
        assert not (bytes_after and not nitems)
        if bytes_after:
            raise RuntimeError("reserved %i bytes for buffer, but data is bigger by %i bytes!" % (buffer_size, bytes_after))
        assert actual_format > 0
        #unlike XGetProperty, we don't need to special case 64-bit:
        cdef int bytes_per_item = actual_format // 8
        cdef int nbytes = bytes_per_item * nitems
        data = (<char *> prop)[:nbytes]
        XFree(prop)
        prop_type = self.get_atom_name(xactual_type)
        log("hex=%s (type=%s, nitems=%i, bytes per item=%i, actual format=%i)",
            hexstr(data), prop_type, nitems, bytes_per_item, actual_format)
        fmt = None
        if prop_type=="INTEGER":
            fmt = {
                8   : b"b",
                16  : b"h",
                32  : b"i",
            }.get(actual_format)
        elif prop_type=="CARDINAL":
            fmt = {
                8   : b"B",
                16  : b"H",
                32  : b"I",
            }.get(actual_format)
        elif prop_type=="FLOAT":
            fmt = b"f"
        if fmt:
            value = struct.unpack(fmt*nitems, data)
            if nitems==1:
                return value[0]
            return value
        return data

    cdef dict get_class_info(self, XIAnyClassInfo *class_info):
        cdef int i
        cdef XIButtonClassInfo *button
        cdef XIKeyClassInfo *key
        cdef XIValuatorClassInfo *valuator
        cdef XIScrollClassInfo *scroll
        cdef XITouchClassInfo *touch
        info = {
            "type"      : CLASS_INFO.get(class_info.type, "unknown type: %i" % class_info.type),
            "sourceid"  : class_info.sourceid,
        }
        if class_info.type==XIButtonClass:
            button = <XIButtonClassInfo*> class_info
            buttons = []
            for i in range(button.num_buttons):
                if button.labels[i]>0:
                    buttons.append(self.get_atom_name(button.labels[i]))
            info["buttons"] = buttons
            #XIButtonState state
        elif class_info.type==XIKeyClass:
            key = <XIKeyClassInfo*> class_info
            keys = []
            for i in range(key.num_keycodes):
                keys.append(key.keycodes[i])
        elif class_info.type==XIValuatorClass:
            valuator = <XIValuatorClassInfo*> class_info
            info |= {
                "number"    : valuator.number,
                "min"       : valuator.min,
                "max"       : valuator.max,
                "value"     : valuator.value,
                "resolution": valuator.resolution,
                "mode"      : valuator.mode,
            }
            if valuator.label:
                info["label"] = self.get_atom_name(valuator.label)
        elif class_info.type==XIScrollClass:
            scroll = <XIScrollClassInfo*> class_info
            info |= {
                "number"        : scroll.number,
                "scroll-type"   : scroll.scroll_type,
                "increment"     : scroll.increment,
                "flags"         : scroll.flags,
            }
        elif class_info.type==XITouchClass:
            touch = <XITouchClassInfo*> class_info
            info |= {
                "mode"          : touch.mode,
                "num-touches"   : touch.num_touches,
            }
        return info

    def gdk_inject(self) -> None:
        self.get_xi_opcode()
        #log.info("XInput Devices:")
        #from xpra.util.str_fn import print_nested_dict
        #print_nested_dict(self.get_devices(), print_fn=log.info)
        self.register_parser()
        self.register_gdk_events()
        #self.select_xi2_events()


cdef X11CoreBindingsInstance singleton = None


def X11XI2Bindings() -> X11CoreBindingsInstance:
    global singleton
    if singleton is None:
        singleton = X11XI2BindingsInstance()
    return singleton



# This file is part of Xpra.
# Copyright (C) 2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.gtk.dialogs.base_gui_window import BaseGUIWindow
from xpra.gtk.configure.common import run_gui
from xpra.util.config import update_config_attribute, with_config
from xpra.gtk.widget import label
from xpra.os_util import gi_import
from xpra.log import Logger

Gtk = gi_import("Gtk")
GLib = gi_import("GLib")

log = Logger("util")


def adj1_100(value: int) -> Gtk.Adjustment:
    return Gtk.Adjustment(value=value, lower=1, upper=100, step_increment=5, page_increment=0, page_size=0)


def make_scale(adjust, marks: dict) -> Gtk.Scale:
    scale = Gtk.Scale(orientation=Gtk.Orientation.HORIZONTAL, adjustment=adjust)
    for value, text in marks.items():
        scale.add_mark(value=value, position=Gtk.PositionType.TOP, markup=text)
    scale.set_digits(0)
    scale.set_draw_value(True)
    scale.set_has_origin(True)
    return scale


class ConfigureGUI(BaseGUIWindow):

    def __init__(self, parent: Gtk.Window | None = None):
        super().__init__(
            "Configure Xpra's Picture Compression",
            "encoding.png",
            wm_class=("xpra-configure-encodings-gui", "Xpra Configure Encodings GUI"),
            default_size=(640, 500),
            header_bar=(False, False),
            parent=parent,
        )

    def populate(self):
        with_config(self.do_populate)

    def do_populate(self, config) -> bool:
        self.clear_vbox()
        self.add_widget(label("Configure Xpra's Picture Compression", font="sans 20"))
        url = "https://github.com/Xpra-org/xpra/blob/master/docs/Usage/Encodings.md#tuning"
        self.add_text_lines((
            f"Please read <a href='{url}'>the documentation</a>.",
        ))
        self.add_widget(Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL))
        self.add_widget(label("Minimum Speed (percentage)", font="Sans 14"))
        self.add_widget(label("Increasing the speed costs bandwidth and CPU time"))
        scale = make_scale(adj1_100(config.min_speed), {
            1: "Lowest",
            50: "Average",
            100: "Highest",
        })
        scale.connect("value-changed", self.speed_changed)
        self.add_widget(scale)
        self.add_widget(Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL))
        self.add_widget(label("Minimum Quality (percentage)", font="Sans 14"))
        self.add_widget(label("Increasing the quality costs bandwidth, CPU time and may also increase the latency"))
        scale = make_scale(adj1_100(config.min_quality), {
            1: "Lowest",
            50: "Average",
            100: "Highest",
        })
        scale.connect("value-changed", self.quality_changed)
        self.add_widget(scale)
        self.add_widget(Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL))
        self.add_widget(label("Auto Refresh Delay (milliseconds)", font="Sans 14"))
        self.add_widget(label("Longer delays may become noticeable but save bandwidth"))
        adjust = Gtk.Adjustment(value=round(config.auto_refresh_delay * 1000), lower=0, upper=1000,
                                step_increment=50, page_increment=0, page_size=0)
        scale = make_scale(adjust, {
            0: "Disabled",
            50: "Fast",
            500: "Average",
            1000: "Slow",
        })
        scale.connect("value-changed", self.ar_changed)
        self.add_widget(scale)
        self.add_widget(Gtk.Separator(orientation=Gtk.Orientation.HORIZONTAL))
        grid = Gtk.Grid()
        grid.set_margin_start(40)
        grid.set_margin_end(40)
        grid.set_row_homogeneous(True)
        grid.set_column_homogeneous(False)
        self.add_widget(grid)
        switch = Gtk.Switch()
        switch.set_state(config.encoding == "grayscale")
        switch.connect("state-set", self.toggle_grayscale)
        lbl = label("Grayscale Mode")
        lbl.set_hexpand(True)
        grid.attach(lbl, 0, 0, 1, 1)
        grid.attach(switch, 1, 0, 1, 1)
        self.show_all()
        return False

    @staticmethod
    def speed_changed(widget) -> None:
        update_config_attribute("min-speed", int(widget.get_value()))

    @staticmethod
    def quality_changed(widget) -> None:
        update_config_attribute("min-quality", int(widget.get_value()))

    @staticmethod
    def ar_changed(widget) -> None:
        update_config_attribute("auto-refresh-delay", int(widget.get_value())/1000)

    @staticmethod
    def toggle_grayscale(_widget, state) -> None:
        update_config_attribute("encoding", "grayscale" if state else "auto")


def main(_args) -> int:
    return run_gui(ConfigureGUI)


if __name__ == "__main__":
    import sys
    sys.exit(main(sys.argv[1:]))
